import json
from typing import TYPE_CHECKING, Any, AsyncGenerator, Dict, List, Optional, Union
from uuid import uuid4

from fastapi import (
    APIRouter,
    BackgroundTasks,
    Depends,
    Form,
    HTTPException,
    Query,
    Request,
    WebSocket,
)
from fastapi.responses import JSONResponse, StreamingResponse
from pydantic import BaseModel

from agno.db.base import BaseDb
from agno.exceptions import InputCheckError, OutputCheckError
from agno.os.auth import (
    get_auth_token_from_request,
    get_authentication_dependency,
    require_resource_access,
    validate_websocket_token,
)
from agno.os.managers import event_buffer, websocket_manager
from agno.os.routers.workflows.schema import WorkflowResponse
from agno.os.schema import (
    BadRequestResponse,
    InternalServerErrorResponse,
    NotFoundResponse,
    UnauthenticatedResponse,
    ValidationErrorResponse,
    WorkflowSummaryResponse,
)
from agno.os.settings import AgnoAPISettings
from agno.os.utils import (
    format_sse_event,
    get_request_kwargs,
    get_workflow_by_id,
)
from agno.run.base import RunStatus
from agno.run.workflow import WorkflowErrorEvent
from agno.utils.log import log_debug, log_warning, logger
from agno.utils.serialize import json_serializer
from agno.workflow.remote import RemoteWorkflow
from agno.workflow.workflow import Workflow

if TYPE_CHECKING:
    from agno.os.app import AgentOS


async def handle_workflow_via_websocket(websocket: WebSocket, message: dict, os: "AgentOS"):
    """Handle workflow execution directly via WebSocket"""
    try:
        workflow_id = message.get("workflow_id")
        session_id = message.get("session_id")
        user_message = message.get("message", "")
        user_id = message.get("user_id")

        if not workflow_id:
            await websocket.send_text(json.dumps({"event": "error", "error": "workflow_id is required"}))
            return

        # Get workflow from OS
        workflow = get_workflow_by_id(
            workflow_id=workflow_id, workflows=os.workflows, db=os.db, registry=os.registry, create_fresh=True
        )
        if not workflow:
            await websocket.send_text(json.dumps({"event": "error", "error": f"Workflow {workflow_id} not found"}))
            return

        if isinstance(workflow, RemoteWorkflow):
            await websocket.send_text(
                json.dumps({"event": "error", "error": "Remote workflows are not supported via WebSocket"})
            )
            return

        # Generate session_id if not provided
        # Use workflow's default session_id if not provided in message
        if not session_id:
            if workflow.session_id:
                session_id = workflow.session_id
            else:
                session_id = str(uuid4())

        # Execute workflow in background with streaming
        await workflow.arun(  # type: ignore
            input=user_message,
            session_id=session_id,
            user_id=user_id,
            stream=True,
            stream_events=True,
            background=True,
            websocket=websocket,
        )

        # NOTE: Don't register the original websocket in the manager
        # It's already handled by the WebSocketHandler passed to the workflow
        # The manager is ONLY for reconnected clients (see handle_workflow_subscription)

    except (InputCheckError, OutputCheckError) as e:
        await websocket.send_text(
            json.dumps(
                {
                    "event": "error",
                    "error": str(e),
                    "error_type": e.type,
                    "error_id": e.error_id,
                    "additional_data": e.additional_data,
                }
            )
        )
    except Exception as e:
        logger.error(f"Error executing workflow via WebSocket: {e}")
        error_payload = {
            "event": "error",
            "error": str(e),
            "error_type": e.type if hasattr(e, "type") else None,
            "error_id": e.error_id if hasattr(e, "error_id") else None,
        }
        error_payload = {k: v for k, v in error_payload.items() if v is not None}
        await websocket.send_text(json.dumps(error_payload))


async def handle_workflow_subscription(websocket: WebSocket, message: dict, os: "AgentOS"):
    """
    Handle subscription/reconnection to an existing workflow run.

    Allows clients to reconnect after page refresh or disconnection and catch up on missed events.
    """
    try:
        run_id = message.get("run_id")
        workflow_id = message.get("workflow_id")
        session_id = message.get("session_id")
        last_event_index = message.get("last_event_index")  # 0-based index of last received event

        if not run_id:
            await websocket.send_text(json.dumps({"event": "error", "error": "run_id is required for subscription"}))
            return

        # Check if run exists in event buffer
        buffer_status = event_buffer.get_run_status(run_id)

        if buffer_status is None:
            # Run not in buffer - check database
            if workflow_id and session_id:
                workflow = get_workflow_by_id(
                    workflow_id=workflow_id, workflows=os.workflows, db=os.db, registry=os.registry, create_fresh=True
                )
                if workflow and isinstance(workflow, Workflow):
                    workflow_run = await workflow.aget_run_output(run_id, session_id)

                    if workflow_run:
                        # Run exists in DB - send all events from DB
                        if workflow_run.events:
                            await websocket.send_text(
                                json.dumps(
                                    {
                                        "event": "replay",
                                        "run_id": run_id,
                                        "status": workflow_run.status.value if workflow_run.status else "unknown",
                                        "total_events": len(workflow_run.events),
                                        "message": "Run completed. Replaying all events from database.",
                                    }
                                )
                            )

                            # Send events one by one
                            for idx, event in enumerate(workflow_run.events):
                                # Convert event to dict and add event_index
                                event_dict = event.model_dump() if hasattr(event, "model_dump") else event.to_dict()
                                event_dict["event_index"] = idx
                                if "run_id" not in event_dict:
                                    event_dict["run_id"] = run_id

                                await websocket.send_text(json.dumps(event_dict, default=json_serializer))
                        else:
                            await websocket.send_text(
                                json.dumps(
                                    {
                                        "event": "replay",
                                        "run_id": run_id,
                                        "status": workflow_run.status.value if workflow_run.status else "unknown",
                                        "total_events": 0,
                                        "message": "Run completed but no events stored.",
                                    }
                                )
                            )
                        return

            # Run not found anywhere
            await websocket.send_text(
                json.dumps({"event": "error", "error": f"Run {run_id} not found in buffer or database"})
            )
            return

        # Run is in buffer (still active or recently completed)
        if buffer_status in [RunStatus.completed, RunStatus.error, RunStatus.cancelled]:
            # Run finished - send all events from buffer
            all_events = event_buffer.get_events(run_id, last_event_index=None)

            await websocket.send_text(
                json.dumps(
                    {
                        "event": "replay",
                        "run_id": run_id,
                        "status": buffer_status.value,
                        "total_events": len(all_events),
                        "message": f"Run {buffer_status.value}. Replaying all events.",
                    }
                )
            )

            # Send all events
            for idx, buffered_event in enumerate(all_events):
                # Convert event to dict and add event_index
                event_dict = (
                    buffered_event.model_dump() if hasattr(buffered_event, "model_dump") else buffered_event.to_dict()
                )
                event_dict["event_index"] = idx
                if "run_id" not in event_dict:
                    event_dict["run_id"] = run_id

                await websocket.send_text(json.dumps(event_dict))
            return

        # Run is still active - send missed events and subscribe to new ones
        missed_events = event_buffer.get_events(run_id, last_event_index)
        current_event_count = event_buffer.get_event_count(run_id)

        if missed_events:
            # Send catch-up notification
            await websocket.send_text(
                json.dumps(
                    {
                        "event": "catch_up",
                        "run_id": run_id,
                        "status": "running",
                        "missed_events": len(missed_events),
                        "current_event_count": current_event_count,
                        "message": f"Catching up on {len(missed_events)} missed events.",
                    }
                )
            )

            # Send missed events
            start_index = (last_event_index + 1) if last_event_index is not None else 0
            for idx, buffered_event in enumerate(missed_events):
                # Convert event to dict and add event_index
                event_dict = (
                    buffered_event.model_dump() if hasattr(buffered_event, "model_dump") else buffered_event.to_dict()
                )
                event_dict["event_index"] = start_index + idx
                if "run_id" not in event_dict:
                    event_dict["run_id"] = run_id

                await websocket.send_text(json.dumps(event_dict))

        # Register websocket for future events
        await websocket_manager.register_websocket(run_id, websocket)

        # Send subscription confirmation
        await websocket.send_text(
            json.dumps(
                {
                    "event": "subscribed",
                    "run_id": run_id,
                    "status": "running",
                    "current_event_count": current_event_count,
                    "message": "Subscribed to workflow run. You will receive new events as they occur.",
                }
            )
        )

        log_debug(f"Client subscribed to workflow run {run_id} (last_event_index: {last_event_index})")

    except Exception as e:
        logger.error(f"Error handling workflow subscription: {e}")
        await websocket.send_text(
            json.dumps(
                {
                    "event": "error",
                    "error": f"Subscription failed: {str(e)}",
                }
            )
        )


async def workflow_response_streamer(
    workflow: Union[Workflow, RemoteWorkflow],
    input: Union[str, Dict[str, Any], List[Any], BaseModel],
    session_id: Optional[str] = None,
    user_id: Optional[str] = None,
    background_tasks: Optional[BackgroundTasks] = None,
    auth_token: Optional[str] = None,
    **kwargs: Any,
) -> AsyncGenerator:
    try:
        # Pass background_tasks if provided
        if background_tasks is not None:
            kwargs["background_tasks"] = background_tasks

        if "stream_events" in kwargs:
            stream_events = kwargs.pop("stream_events")
        else:
            stream_events = True

        # Pass auth_token for remote workflows
        if auth_token and isinstance(workflow, RemoteWorkflow):
            kwargs["auth_token"] = auth_token

        run_response = workflow.arun(  # type: ignore
            input=input,
            session_id=session_id,
            user_id=user_id,
            stream=True,
            stream_events=stream_events,
            **kwargs,
        )

        async for run_response_chunk in run_response:
            yield format_sse_event(run_response_chunk)  # type: ignore

    except (InputCheckError, OutputCheckError) as e:
        error_response = WorkflowErrorEvent(
            error=str(e),
            error_type=e.type,
            error_id=e.error_id,
            additional_data=e.additional_data,
        )
        yield format_sse_event(error_response)

    except Exception as e:
        import traceback

        traceback.print_exc()
        error_response = WorkflowErrorEvent(
            error=str(e),
            error_type=e.type if hasattr(e, "type") else None,
            error_id=e.error_id if hasattr(e, "error_id") else None,
        )
        yield format_sse_event(error_response)
        return


def get_websocket_router(
    os: "AgentOS",
    settings: AgnoAPISettings = AgnoAPISettings(),
) -> APIRouter:
    """
    Create WebSocket router with support for both legacy (os_security_key) and JWT authentication.

    WebSocket endpoints handle authentication internally via message-based auth.
    Authentication methods (in order of precedence):
    1. JWT tokens - if JWTMiddleware is configured (via app.state.jwt_middleware)
    2. Legacy bearer token - if settings.os_security_key is set
    3. No authentication - if neither is configured

    The JWT middleware instance is accessed from app.state.jwt_middleware, which is set
    by AgentOS when authorization is enabled. This allows reusing the same validation
    logic and loaded keys as the HTTP middleware.

    Args:
        os: The AgentOS instance
        settings: API settings (includes os_security_key for legacy auth)
    """
    ws_router = APIRouter()

    @ws_router.websocket(
        "/workflows/ws",
        name="workflow_websocket",
    )
    async def workflow_websocket_endpoint(websocket: WebSocket):
        """WebSocket endpoint for receiving real-time workflow events"""
        # Check if JWT validator is configured (set by AgentOS when authorization=True)
        jwt_validator = getattr(websocket.app.state, "jwt_validator", None)
        jwt_auth_enabled = jwt_validator is not None

        # Determine auth requirements - JWT takes precedence over legacy
        requires_auth = jwt_auth_enabled or bool(settings.os_security_key)

        await websocket_manager.connect(websocket, requires_auth=requires_auth)

        # Store user context from JWT auth
        websocket_user_context: Dict[str, Any] = {}

        try:
            while True:
                data = await websocket.receive_text()
                message = json.loads(data)
                action = message.get("action")

                # Handle authentication first
                if action == "authenticate":
                    token = message.get("token")
                    if not token:
                        await websocket.send_text(json.dumps({"event": "auth_error", "error": "Token is required"}))
                        continue

                    if jwt_auth_enabled and jwt_validator:
                        # Use JWT validator for token validation
                        try:
                            payload = jwt_validator.validate_token(token)
                            claims = jwt_validator.extract_claims(payload)
                            await websocket_manager.authenticate_websocket(websocket)

                            # Store user context from JWT
                            websocket_user_context["user_id"] = claims["user_id"]
                            websocket_user_context["scopes"] = claims["scopes"]
                            websocket_user_context["payload"] = payload

                            # Include user info in auth success message
                            await websocket.send_text(
                                json.dumps(
                                    {
                                        "event": "authenticated",
                                        "message": "JWT authentication successful.",
                                        "user_id": claims["user_id"],
                                    }
                                )
                            )
                        except Exception as e:
                            error_msg = str(e) if str(e) else "Invalid token"
                            error_type = "expired" if "expired" in error_msg.lower() else "invalid_token"
                            await websocket.send_text(
                                json.dumps(
                                    {
                                        "event": "auth_error",
                                        "error": error_msg,
                                        "error_type": error_type,
                                    }
                                )
                            )
                        continue
                    elif validate_websocket_token(token, settings):
                        # Legacy os_security_key authentication
                        await websocket_manager.authenticate_websocket(websocket)
                    else:
                        await websocket.send_text(json.dumps({"event": "auth_error", "error": "Invalid token"}))
                    continue

                # Check authentication for all other actions (only when required)
                elif requires_auth and not websocket_manager.is_authenticated(websocket):
                    auth_type = "JWT" if jwt_auth_enabled else "bearer token"
                    await websocket.send_text(
                        json.dumps(
                            {
                                "event": "auth_required",
                                "error": f"Authentication required. Send authenticate action with valid {auth_type}.",
                            }
                        )
                    )
                    continue

                # Handle authenticated actions
                elif action == "ping":
                    await websocket.send_text(json.dumps({"event": "pong"}))

                elif action == "start-workflow":
                    # Add user context to message if available from JWT auth
                    if websocket_user_context:
                        if "user_id" not in message and websocket_user_context.get("user_id"):
                            message["user_id"] = websocket_user_context["user_id"]
                    # Handle workflow execution directly via WebSocket
                    await handle_workflow_via_websocket(websocket, message, os)

                elif action == "reconnect":
                    # Subscribe/reconnect to an existing workflow run
                    await handle_workflow_subscription(websocket, message, os)

                else:
                    await websocket.send_text(json.dumps({"event": "error", "error": f"Unknown action: {action}"}))

        except Exception as e:
            if "1012" not in str(e) and "1001" not in str(e):
                logger.error(f"WebSocket error: {e}")
        finally:
            # Clean up the websocket connection
            await websocket_manager.disconnect_websocket(websocket)

    return ws_router


def get_workflow_router(
    os: "AgentOS",
    settings: AgnoAPISettings = AgnoAPISettings(),
) -> APIRouter:
    """Create the workflow router with comprehensive OpenAPI documentation."""
    router = APIRouter(
        dependencies=[Depends(get_authentication_dependency(settings))],
        responses={
            400: {"description": "Bad Request", "model": BadRequestResponse},
            401: {"description": "Unauthorized", "model": UnauthenticatedResponse},
            404: {"description": "Not Found", "model": NotFoundResponse},
            422: {"description": "Validation Error", "model": ValidationErrorResponse},
            500: {"description": "Internal Server Error", "model": InternalServerErrorResponse},
        },
    )

    @router.get(
        "/workflows",
        response_model=List[WorkflowSummaryResponse],
        response_model_exclude_none=True,
        tags=["Workflows"],
        operation_id="get_workflows",
        summary="List All Workflows",
        description=(
            "Retrieve a comprehensive list of all workflows configured in this OS instance.\n\n"
            "**Return Information:**\n"
            "- Workflow metadata (ID, name, description)\n"
            "- Input schema requirements\n"
            "- Step sequence and execution flow\n"
            "- Associated agents and teams"
        ),
        responses={
            200: {
                "description": "List of workflows retrieved successfully",
                "content": {
                    "application/json": {
                        "example": [
                            {
                                "id": "content-creation-workflow",
                                "name": "Content Creation Workflow",
                                "description": "Automated content creation from blog posts to social media",
                                "db_id": "123",
                            }
                        ]
                    }
                },
            }
        },
    )
    async def get_workflows(request: Request) -> List[WorkflowSummaryResponse]:
        # Filter workflows based on user's scopes (only if authorization is enabled)
        if getattr(request.state, "authorization_enabled", False):
            from agno.os.auth import filter_resources_by_access, get_accessible_resources

            # Check if user has any workflow scopes at all
            accessible_ids = get_accessible_resources(request, "workflows")
            if not accessible_ids:
                raise HTTPException(status_code=403, detail="Insufficient permissions")

            accessible_workflows = filter_resources_by_access(request, os.workflows or [], "workflows")
        else:
            accessible_workflows = os.workflows or []

        workflows: List[WorkflowSummaryResponse] = []
        if accessible_workflows:
            for workflow in accessible_workflows:
                workflows.append(WorkflowSummaryResponse.from_workflow(workflow=workflow, is_component=False))

        if os.db and isinstance(os.db, BaseDb):
            from agno.workflow.workflow import get_workflows

            for db_workflow in get_workflows(db=os.db, registry=os.registry):
                try:
                    workflows.append(WorkflowSummaryResponse.from_workflow(workflow=db_workflow, is_component=True))
                except Exception as e:
                    workflow_id = getattr(db_workflow, "id", "unknown")
                    logger.error(f"Error converting workflow {workflow_id} to response: {e}")
                    continue

        return workflows

    @router.get(
        "/workflows/{workflow_id}",
        response_model=WorkflowResponse,
        response_model_exclude_none=True,
        tags=["Workflows"],
        operation_id="get_workflow",
        summary="Get Workflow Details",
        description=("Retrieve detailed configuration and step information for a specific workflow."),
        responses={
            200: {
                "description": "Workflow details retrieved successfully",
                "content": {
                    "application/json": {
                        "example": {
                            "id": "content-creation-workflow",
                            "name": "Content Creation Workflow",
                            "description": "Automated content creation from blog posts to social media",
                            "db_id": "123",
                        }
                    }
                },
            },
            404: {"description": "Workflow not found", "model": NotFoundResponse},
        },
        dependencies=[Depends(require_resource_access("workflows", "read", "workflow_id"))],
    )
    async def get_workflow(
        workflow_id: str,
        request: Request,
        version: Optional[int] = Query(None, description="Workflow version to retrieve"),
    ) -> WorkflowResponse:
        workflow = get_workflow_by_id(
            workflow_id=workflow_id,
            workflows=os.workflows,
            db=os.db,
            version=version,
            registry=os.registry,
            create_fresh=True,
        )
        if workflow is None:
            raise HTTPException(status_code=404, detail="Workflow not found")
        if isinstance(workflow, RemoteWorkflow):
            return await workflow.get_workflow_config()
        else:
            return await WorkflowResponse.from_workflow(workflow=workflow)

    @router.post(
        "/workflows/{workflow_id}/runs",
        tags=["Workflows"],
        operation_id="create_workflow_run",
        response_model_exclude_none=True,
        summary="Execute Workflow",
        description=(
            "Execute a workflow with the provided input data. Workflows can run in streaming or batch mode.\n\n"
            "**Execution Modes:**\n"
            "- **Streaming (`stream=true`)**: Real-time step-by-step execution updates via SSE\n"
            "- **Non-Streaming (`stream=false`)**: Complete workflow execution with final result\n\n"
            "**Workflow Execution Process:**\n"
            "1. Input validation against workflow schema\n"
            "2. Sequential or parallel step execution based on workflow design\n"
            "3. Data flow between steps with transformation\n"
            "4. Error handling and automatic retries where configured\n"
            "5. Final result compilation and response\n\n"
            "**Session Management:**\n"
            "Workflows support session continuity for stateful execution across multiple runs."
        ),
        responses={
            200: {
                "description": "Workflow executed successfully",
                "content": {
                    "text/event-stream": {
                        "example": 'event: RunStarted\ndata: {"content": "Hello!", "run_id": "123..."}\n\n'
                    },
                },
            },
            400: {"description": "Invalid input data or workflow configuration", "model": BadRequestResponse},
            404: {"description": "Workflow not found", "model": NotFoundResponse},
            500: {"description": "Workflow execution error", "model": InternalServerErrorResponse},
        },
        dependencies=[Depends(require_resource_access("workflows", "run", "workflow_id"))],
    )
    async def create_workflow_run(
        workflow_id: str,
        request: Request,
        background_tasks: BackgroundTasks,
        message: str = Form(..., description="The input message or prompt to send to the workflow"),
        stream: bool = Form(True, description="Enable streaming responses via Server-Sent Events (SSE)"),
        session_id: Optional[str] = Form(
            None, description="Session ID for conversation continuity. If not provided, a new session is created"
        ),
        user_id: Optional[str] = Form(None, description="User identifier for tracking and personalization"),
        version: Optional[int] = Form(None, description="Workflow version to use for this run"),
    ):
        kwargs = await get_request_kwargs(request, create_workflow_run)

        if hasattr(request.state, "user_id") and request.state.user_id is not None:
            if user_id and user_id != request.state.user_id:
                log_warning("User ID parameter passed in both request state and kwargs, using request state")
            user_id = request.state.user_id
        if hasattr(request.state, "session_id") and request.state.session_id is not None:
            if session_id and session_id != request.state.session_id:
                log_warning("Session ID parameter passed in both request state and kwargs, using request state")
            session_id = request.state.session_id
        if hasattr(request.state, "session_state") and request.state.session_state is not None:
            session_state = request.state.session_state
            if "session_state" in kwargs:
                log_warning("Session state parameter passed in both request state and kwargs, using request state")
            kwargs["session_state"] = session_state
        if hasattr(request.state, "dependencies") and request.state.dependencies is not None:
            dependencies = request.state.dependencies
            if "dependencies" in kwargs:
                log_warning("Dependencies parameter passed in both request state and kwargs, using request state")
            kwargs["dependencies"] = dependencies
        if hasattr(request.state, "metadata") and request.state.metadata is not None:
            metadata = request.state.metadata
            if "metadata" in kwargs:
                log_warning("Metadata parameter passed in both request state and kwargs, using request state")
            kwargs["metadata"] = metadata

        # Retrieve the workflow by ID
        workflow = get_workflow_by_id(
            workflow_id=workflow_id,
            workflows=os.workflows,
            db=os.db,
            version=version,
            registry=os.registry,
            create_fresh=True,
        )
        if workflow is None:
            raise HTTPException(status_code=404, detail="Workflow not found")

        if session_id:
            logger.debug(f"Continuing session: {session_id}")
        else:
            logger.debug("Creating new session")
            session_id = str(uuid4())

        # Extract auth token for remote workflows
        auth_token = get_auth_token_from_request(request)

        # Return based on stream parameter
        try:
            if stream:
                return StreamingResponse(
                    workflow_response_streamer(
                        workflow,
                        input=message,
                        session_id=session_id,
                        user_id=user_id,
                        background_tasks=background_tasks,
                        auth_token=auth_token,
                        **kwargs,
                    ),
                    media_type="text/event-stream",
                )
            else:
                # Pass auth_token for remote workflows
                if auth_token and isinstance(workflow, RemoteWorkflow):
                    kwargs["auth_token"] = auth_token

                run_response = await workflow.arun(
                    input=message,
                    session_id=session_id,
                    user_id=user_id,
                    stream=False,
                    background_tasks=background_tasks,
                    **kwargs,
                )
                return run_response.to_dict()

        except InputCheckError as e:
            raise HTTPException(status_code=400, detail=str(e))
        except Exception as e:
            # Handle unexpected runtime errors
            raise HTTPException(status_code=500, detail=f"Error running workflow: {str(e)}")

    @router.post(
        "/workflows/{workflow_id}/runs/{run_id}/cancel",
        tags=["Workflows"],
        operation_id="cancel_workflow_run",
        summary="Cancel Workflow Run",
        description=(
            "Cancel a currently executing workflow run, stopping all active steps and cleanup.\n"
            "**Note:** Complex workflows with multiple parallel steps may take time to fully cancel."
        ),
        responses={
            200: {},
            404: {"description": "Workflow or run not found", "model": NotFoundResponse},
            500: {"description": "Failed to cancel workflow run", "model": InternalServerErrorResponse},
        },
        dependencies=[Depends(require_resource_access("workflows", "run", "workflow_id"))],
    )
    async def cancel_workflow_run(workflow_id: str, run_id: str):
        workflow = get_workflow_by_id(
            workflow_id=workflow_id, workflows=os.workflows, db=os.db, registry=os.registry, create_fresh=True
        )

        if workflow is None:
            raise HTTPException(status_code=404, detail="Workflow not found")

        # cancel_run always stores cancellation intent (even for not-yet-registered runs
        # in cancel-before-start scenarios), so we always return success.
        await workflow.acancel_run(run_id=run_id)
        return JSONResponse(content={}, status_code=200)

    @router.get(
        "/workflows/{workflow_id}/runs/{run_id}",
        tags=["Workflows"],
        operation_id="get_workflow_run",
        summary="Get Workflow Run",
        description=(
            "Retrieve the status and output of a workflow run. Use this to poll for run completion.\n\n"
            "Requires the `session_id` that was returned when the run was created."
        ),
        responses={
            200: {"description": "Run output retrieved successfully"},
            404: {"description": "Workflow or run not found", "model": NotFoundResponse},
        },
        dependencies=[Depends(require_resource_access("workflows", "run", "workflow_id"))],
    )
    async def get_workflow_run(
        workflow_id: str,
        run_id: str,
        session_id: str = Query(..., description="Session ID for the run"),
    ):
        workflow = get_workflow_by_id(
            workflow_id=workflow_id, workflows=os.workflows, db=os.db, registry=os.registry, create_fresh=True
        )
        if workflow is None:
            raise HTTPException(status_code=404, detail="Workflow not found")
        if isinstance(workflow, RemoteWorkflow):
            raise HTTPException(status_code=400, detail="Run polling is not supported for remote workflows")

        run_output = await workflow.aget_run_output(run_id=run_id, session_id=session_id)
        if run_output is None:
            raise HTTPException(status_code=404, detail="Run not found")

        return run_output.to_dict()

    return router



import io
import mimetypes
import os
from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Union

import httpx

from agno.utils.audio import pcm_to_wav_bytes
from agno.utils.log import log_error, log_info, log_warning
from agno.utils.tokens import get_image_type

_BASE_URL = "https://graph.facebook.com"
_API_VERSION = "v22.0"


@dataclass
class WhatsAppConfig:
    # Resolved once at startup by attach_routes; passed to all helpers
    access_token: str
    phone_number_id: str
    verify_token: Optional[str] = None
    # Timeout in seconds for media downloads/uploads; configurable via Whatsapp(media_timeout=)
    media_timeout: int = 30

    @classmethod
    def init(
        cls,
        access_token: Optional[str] = None,
        phone_number_id: Optional[str] = None,
        verify_token: Optional[str] = None,
        media_timeout: int = 30,
    ) -> "WhatsAppConfig":
        token = access_token or os.getenv("WHATSAPP_ACCESS_TOKEN")
        phone_id = phone_number_id or os.getenv("WHATSAPP_PHONE_NUMBER_ID")
        v_token = verify_token or os.getenv("WHATSAPP_VERIFY_TOKEN")
        if not token:
            raise ValueError("WHATSAPP_ACCESS_TOKEN is not set. Set the environment variable or pass access_token.")
        if not phone_id:
            raise ValueError(
                "WHATSAPP_PHONE_NUMBER_ID is not set. Set the environment variable or pass phone_number_id."
            )
        return cls(access_token=token, phone_number_id=phone_id, verify_token=v_token, media_timeout=media_timeout)

    def messages_url(self) -> str:
        return f"{_BASE_URL}/{_API_VERSION}/{self.phone_number_id}/messages"

    def media_url(self) -> str:
        return f"{_BASE_URL}/{_API_VERSION}/{self.phone_number_id}/media"

    def auth_headers(self) -> dict:
        return {"Authorization": f"Bearer {self.access_token}"}


@dataclass
class MessageContent:
    text: str
    image_id: Optional[str] = None
    video_id: Optional[str] = None
    audio_id: Optional[str] = None
    doc_id: Optional[str] = None


def extract_message_content(message: dict) -> Optional[MessageContent]:
    msg_type = message.get("type")

    if msg_type == "text":
        text = message["text"]["body"]
        log_info(text)
        return MessageContent(text=text)

    if msg_type == "image":
        return MessageContent(
            text=message.get("image", {}).get("caption", ""),
            image_id=message["image"]["id"],
        )

    if msg_type == "video":
        return MessageContent(
            text=message.get("video", {}).get("caption", ""),
            video_id=message["video"]["id"],
        )

    if msg_type == "audio":
        return MessageContent(text="", audio_id=message["audio"]["id"])

    if msg_type == "document":
        return MessageContent(
            text=message.get("document", {}).get("caption", ""),
            doc_id=message["document"]["id"],
        )

    # Interactive replies carry the selected option's title and description
    if msg_type == "interactive":
        interactive = message.get("interactive", {})
        interactive_type = interactive.get("type")
        if interactive_type == "button_reply":
            reply = interactive.get("button_reply", {})
            text = reply.get("title", "")
            log_info(f"Button reply: id={reply.get('id')} title={text}")
            return MessageContent(text=text)
        if interactive_type == "list_reply":
            reply = interactive.get("list_reply", {})
            text = reply.get("title", "")
            description = reply.get("description", "")
            if description:
                text = f"{text}: {description}"
            log_info(f"List reply: id={reply.get('id')} title={text}")
            return MessageContent(text=text)
        log_warning(f"Unknown interactive type: {interactive_type}")
        return None

    log_warning(f"Unknown message type: {msg_type}")
    return None


_WHATSAPP_AUDIO_MIMES = {"audio/aac", "audio/mp4", "audio/mpeg", "audio/amr", "audio/ogg", "audio/wav"}


async def get_media_async(media_id: str, config: WhatsAppConfig) -> Union[dict, bytes]:
    url = f"{_BASE_URL}/{_API_VERSION}/{media_id}"
    headers = config.auth_headers()
    timeout = config.media_timeout

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.get(url, headers=headers)
            response.raise_for_status()
            data = response.json()
        media_url = data.get("url")
    except httpx.HTTPError as e:
        return {"error": str(e)}

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            response = await client.get(media_url, headers=headers)
            response.raise_for_status()
            return response.content
    except httpx.HTTPError as e:
        return {"error": str(e)}


@dataclass
class _MediaResult:
    content: Optional[bytes] = None
    mime_type: Optional[str] = None
    skip_reason: Optional[str] = None


async def _download_media(media_id: str, media_label: str, config: WhatsAppConfig) -> _MediaResult:
    url = f"{_BASE_URL}/{_API_VERSION}/{media_id}"
    headers = config.auth_headers()
    timeout = config.media_timeout
    mime_type: Optional[str] = None

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            resp = await client.get(url, headers=headers)
            resp.raise_for_status()
            metadata = resp.json()
        media_url = metadata.get("url")
        mime_type = metadata.get("mime_type")
    except httpx.HTTPError as e:
        reason = f"{media_label} (metadata fetch failed: {e})"
        log_warning(f"Media download skipped: {reason}")
        return _MediaResult(skip_reason=reason)

    try:
        async with httpx.AsyncClient(timeout=timeout) as client:
            resp = await client.get(media_url, headers=headers)
            resp.raise_for_status()
            return _MediaResult(content=resp.content, mime_type=mime_type)
    except httpx.HTTPError as e:
        reason = f"{media_label} (download failed: {e})"
        log_warning(f"Media download skipped: {reason}")
        return _MediaResult(skip_reason=reason)


_MEDIA_FIELDS = ("image_id", "video_id", "audio_id", "doc_id")
_MEDIA_LABELS = ("image", "video", "audio", "document")


async def download_event_media_async(parsed: "MessageContent", config: WhatsAppConfig) -> Tuple[dict, List[str]]:
    from agno.media import Audio, File, Image, Video

    run_kwargs: dict = {}
    skipped: List[str] = []

    for field_name, label in zip(_MEDIA_FIELDS, _MEDIA_LABELS):
        media_id = getattr(parsed, field_name, None)
        if not media_id:
            continue
        result = await _download_media(media_id, label, config)
        if result.skip_reason:
            skipped.append(result.skip_reason)
            continue
        content = result.content
        mime = result.mime_type
        if label == "image":
            run_kwargs["images"] = [Image(content=content, mime_type=mime)]
        elif label == "video":
            run_kwargs["videos"] = [Video(content=content, mime_type=mime)]
        elif label == "audio":
            run_kwargs["audio"] = [Audio(content=content, mime_type=mime)]
        elif label == "document":
            run_kwargs["files"] = [File(content=content, mime_type=mime)]

    return run_kwargs, skipped


async def upload_media_async(
    media_data: bytes, mime_type: str, filename: str, config: WhatsAppConfig
) -> Union[str, dict]:
    url = config.media_url()
    headers = config.auth_headers()
    data = {"messaging_product": "whatsapp", "type": mime_type}

    try:
        file_data = io.BytesIO(media_data)
        files = {"file": (filename, file_data, mime_type)}
        async with httpx.AsyncClient(timeout=config.media_timeout) as client:
            response = await client.post(url, headers=headers, data=data, files=files)
            response.raise_for_status()
            json_resp = response.json()
            result_id = json_resp.get("id")
            if not result_id:
                return {"error": "Media ID not found in response", "response": json_resp}
            return result_id
    except httpx.HTTPError as e:
        return {"error": str(e)}
    except Exception as e:
        return {"error": str(e)}


async def _send_text(recipient: str, text: str, config: WhatsAppConfig, preview_url: bool = False) -> None:
    url = config.messages_url()
    headers = config.auth_headers()

    data = {
        "messaging_product": "whatsapp",
        "recipient_type": "individual",
        "to": recipient,
        "type": "text",
        "text": {"preview_url": preview_url, "body": text},
    }

    try:
        async with httpx.AsyncClient() as client:
            response = await client.post(url, headers=headers, json=data)
            response.raise_for_status()
    except httpx.HTTPStatusError as e:
        log_error(f"Failed to send WhatsApp text message: {e}")
        log_error(f"Error response: {e.response.text}")
        raise
    except Exception as e:
        log_error(f"Unexpected error sending WhatsApp text message: {str(e)}")
        raise


async def _send_media(
    media_type: str,
    media_id: str,
    recipient: str,
    config: WhatsAppConfig,
    caption: Optional[str] = None,
    filename: Optional[str] = None,
) -> None:
    url = config.messages_url()
    headers = config.auth_headers()

    media_payload: dict = {"id": media_id}
    if caption:
        media_payload["caption"] = caption
    if filename and media_type == "document":
        media_payload["filename"] = filename

    data = {
        "messaging_product": "whatsapp",
        "recipient_type": "individual",
        "to": recipient,
        "type": media_type,
        media_type: media_payload,
    }

    try:
        async with httpx.AsyncClient() as client:
            response = await client.post(url, headers=headers, json=data)
            response.raise_for_status()
    except httpx.HTTPStatusError as e:
        log_error(f"Failed to send WhatsApp {media_type} message: {e}")
        log_error(f"Error response: {e.response.text}")
        raise
    except Exception as e:
        log_error(f"Unexpected error sending WhatsApp {media_type} message: {str(e)}")
        raise


async def typing_indicator_async(message_id: Optional[str], config: WhatsAppConfig) -> Optional[dict]:
    if not message_id:
        return None

    url = config.messages_url()
    headers = config.auth_headers()
    data = {
        "messaging_product": "whatsapp",
        "status": "read",
        "message_id": message_id,
        "typing_indicator": {"type": "text"},
    }

    try:
        async with httpx.AsyncClient() as client:
            response = await client.post(url, headers=headers, json=data)
            response.raise_for_status()
    except Exception as e:
        return {"error": str(e)}
    return None


def format_message(text: str) -> str:
    import re

    # ## Header → *Header* (WhatsApp has no header rendering)
    text = re.sub(r"^#{1,6}\s+(.+)$", r"*\1*", text, flags=re.MULTILINE)
    # **bold** → *bold* (WhatsApp uses single asterisks)
    text = re.sub(r"\*\*(.+?)\*\*", r"*\1*", text)
    # [text](url) → text (url)
    text = re.sub(r"\[([^\]]+)\]\(([^)]+)\)", r"\1 (\2)", text)
    return text


async def send_whatsapp_message_async(
    recipient: str, message: Any, config: WhatsAppConfig, italics: bool = False
) -> None:
    # output_schema responses arrive as pydantic models; coerce to string for WhatsApp
    if message is not None and not isinstance(message, str):
        from pydantic import BaseModel

        message = message.model_dump_json(indent=2) if isinstance(message, BaseModel) else str(message)
    if not message or not message.strip():
        return

    message = format_message(message)

    def _format(text: str) -> str:
        if italics:
            return "\n".join([f"_{line}_" for line in text.split("\n")])
        return text

    # WhatsApp limit is 4096 chars; split at 4000 to leave room for batch prefix
    if len(message) <= 4096:
        await _send_text(recipient=recipient, text=_format(message), config=config)
        return

    message_batches = [message[i : i + 4000] for i in range(0, len(message), 4000)]
    for i, batch in enumerate(message_batches, 1):
        batch_message = f"[{i}/{len(message_batches)}] {batch}"
        await _send_text(recipient=recipient, text=_format(batch_message), config=config)


async def upload_and_send_media_async(
    media_items: list,
    media_type: str,
    recipient: str,
    config: WhatsAppConfig,
    response_content: Optional[str] = None,
    send_text_fallback: bool = True,
) -> bool:
    any_sent = False
    for item in media_items:
        raw_bytes = await item.aget_content_bytes()
        if not raw_bytes:
            log_warning(f"Could not process {media_type} content for user {recipient}. Type: {type(item.content)}")
            if send_text_fallback:
                await send_whatsapp_message_async(recipient, response_content or "", config)
            continue

        if media_type == "image":
            # WhatsApp only accepts image/jpeg and image/png
            detected = get_image_type(raw_bytes)
            if detected in ("jpeg", "png"):
                fmt = detected
            else:
                # GIF, WebP, HEIC, unknown — unsupported by WhatsApp API
                log_warning(f"Unsupported image format '{detected}' for WhatsApp, skipping upload")
                if send_text_fallback and response_content:
                    await _send_text(recipient=recipient, text=response_content, config=config)
                continue
            mime_type = f"image/{fmt}"
            filename = f"image.{fmt}"
        elif media_type == "document":
            filename = item.name or item.filename or "document"
            mime_type = mimetypes.guess_type(filename)[0] or "application/octet-stream"
        elif media_type == "video":
            mime_type = getattr(item, "mime_type", None) or "video/mp4"
            filename = f"video.{mime_type.split('/')[-1]}"
        elif media_type == "audio":
            mime_type = item.mime_type or "audio/mpeg"
            if mime_type.split(";")[0] in _WHATSAPP_AUDIO_MIMES:
                fmt = item.format or mime_type.split("/")[-1]
                filename = f"audio.{fmt}"
            else:
                # Raw PCM (e.g. Gemini TTS "audio/L16;rate=24000") — wrap as WAV
                raw_bytes = pcm_to_wav_bytes(raw_bytes, channels=item.channels, rate=item.sample_rate)
                mime_type, filename = "audio/wav", "audio.wav"
        else:
            mime_type, filename = "application/octet-stream", media_type

        mid = await upload_media_async(media_data=raw_bytes, mime_type=mime_type, filename=filename, config=config)
        if isinstance(mid, dict):
            log_warning(f"{media_type.title()} upload failed for user {recipient}: {mid}")
            if send_text_fallback:
                await send_whatsapp_message_async(recipient, response_content or "", config)
            continue

        # Caption only the first item to avoid repeating text on every media
        caption = None
        if not any_sent and media_type in ("image", "video", "document") and response_content:
            caption = response_content[:1021] + "..." if len(response_content) > 1024 else response_content
        await _send_media(
            media_type=media_type,
            media_id=mid,
            recipient=recipient,
            config=config,
            caption=caption,
            filename=filename if media_type == "document" else None,
        )
        any_sent = True
    return any_sent



import json
from typing import Any, Dict, List, Optional

from agno.tools import Toolkit
from agno.utils.log import log_debug, log_error

try:
    from docling.datamodel.base_models import InputFormat
    from docling.datamodel.pipeline_options import (
        EasyOcrOptions,
        OcrAutoOptions,
        OcrMacOptions,
        PdfPipelineOptions,
        RapidOcrOptions,
        TesseractCliOcrOptions,
        TesseractOcrOptions,
    )
    from docling.document_converter import DocumentConverter, PdfFormatOption
except ImportError:
    raise ImportError("`docling` not installed. Please install using `pip install docling`")


class DoclingTools(Toolkit):
    """
    Toolkit for converting documents with Docling.

    Supports local files and URLs. Export formats: markdown, text, html, html_split_page,
    json, yaml, doctags, and vtt.
    Advanced pipeline/OCR options can be configured via init params.

    PDF/OCR options (init args):
    - pdf_enable_ocr: bool
    - pdf_ocr_engine: "auto" | "easyocr" | "tesseract" | "tesseract_cli" | "ocrmac" | "rapidocr"
    - pdf_ocr_lang: list of language codes
    - pdf_force_full_page_ocr: bool
    - pdf_enable_table_structure: bool
    - pdf_enable_picture_description: bool
    - pdf_enable_picture_classification: bool
    - pdf_document_timeout: float (seconds)
    - pdf_enable_remote_services: bool

    Note:
    Some OCR engines may require additional runtime dependencies. For example,
    `easyocr` can require installing EasyOCR and its model/runtime stack in the
    active environment.
    """

    def __init__(
        self,
        converter: Optional[DocumentConverter] = None,
        max_chars: Optional[int] = None,
        allowed_input_formats: Optional[List[str]] = None,
        format_options: Optional[Dict[Any, Any]] = None,
        pdf_pipeline_options: Optional[PdfPipelineOptions] = None,
        pdf_enable_ocr: Optional[bool] = None,
        pdf_ocr_engine: Optional[str] = None,
        pdf_ocr_lang: Optional[List[str]] = None,
        pdf_force_full_page_ocr: Optional[bool] = None,
        pdf_enable_table_structure: Optional[bool] = None,
        pdf_enable_picture_description: Optional[bool] = None,
        pdf_enable_picture_classification: Optional[bool] = None,
        pdf_document_timeout: Optional[float] = None,
        pdf_enable_remote_services: Optional[bool] = None,
        enable_convert_to_markdown: bool = True,
        enable_convert_to_text: bool = True,
        enable_convert_to_html: bool = True,
        enable_convert_to_html_split_page: bool = True,
        enable_convert_to_json: bool = True,
        enable_convert_to_yaml: bool = True,
        enable_convert_to_doctags: bool = True,
        enable_convert_to_vtt: bool = True,
        enable_convert_string_content: bool = True,
        enable_list_supported_parsers: bool = True,
        all: bool = False,
        **kwargs,
    ):
        self.converter: DocumentConverter = converter or self._build_converter(
            allowed_input_formats=allowed_input_formats,
            format_options=format_options,
            pdf_pipeline_options=pdf_pipeline_options,
            pdf_enable_ocr=pdf_enable_ocr,
            pdf_ocr_engine=pdf_ocr_engine,
            pdf_ocr_lang=pdf_ocr_lang,
            pdf_force_full_page_ocr=pdf_force_full_page_ocr,
            pdf_enable_table_structure=pdf_enable_table_structure,
            pdf_enable_picture_description=pdf_enable_picture_description,
            pdf_enable_picture_classification=pdf_enable_picture_classification,
            pdf_document_timeout=pdf_document_timeout,
            pdf_enable_remote_services=pdf_enable_remote_services,
        )
        self.max_chars = max_chars

        tools: List[Any] = []
        if all or enable_convert_to_markdown:
            tools.append(self.convert_to_markdown)
        if all or enable_convert_to_text:
            tools.append(self.convert_to_text)
        if all or enable_convert_to_html:
            tools.append(self.convert_to_html)
        if all or enable_convert_to_html_split_page:
            tools.append(self.convert_to_html_split_page)
        if all or enable_convert_to_json:
            tools.append(self.convert_to_json)
        if all or enable_convert_to_yaml:
            tools.append(self.convert_to_yaml)
        if all or enable_convert_to_doctags:
            tools.append(self.convert_to_doctags)
        if all or enable_convert_to_vtt:
            tools.append(self.convert_to_vtt)
        if all or enable_convert_string_content:
            tools.append(self.convert_string_content)
        if all or enable_list_supported_parsers:
            tools.append(self.list_supported_parsers)

        super().__init__(name="docling_tools", tools=tools, **kwargs)

    def convert_to_markdown(
        self,
        source: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        """Convert a local file or URL to Markdown format using Docling.

        Args:
            source (str): Local file path or URL to the document.
            headers (Optional[Dict[str, str]]): Optional HTTP headers used for URL fetching.
            raises_on_error (bool): If True, raises conversion exceptions in Docling internals.
            max_num_pages (Optional[int]): Maximum number of pages to process from the source.
            max_file_size (Optional[int]): Maximum file size in bytes allowed for processing.

        Returns:
            str: The converted Markdown content, or an error message if conversion fails.
        """
        return self._convert_and_export(
            source,
            export_format="markdown",
            headers=headers,
            raises_on_error=raises_on_error,
            max_num_pages=max_num_pages,
            max_file_size=max_file_size,
        )

    def convert_to_text(
        self,
        source: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        """Convert a local file or URL to plain text format using Docling.

        Args:
            source (str): Local file path or URL to the document.
            headers (Optional[Dict[str, str]]): Optional HTTP headers used for URL fetching.
            raises_on_error (bool): If True, raises conversion exceptions in Docling internals.
            max_num_pages (Optional[int]): Maximum number of pages to process from the source.
            max_file_size (Optional[int]): Maximum file size in bytes allowed for processing.

        Returns:
            str: The converted plain text content, or an error message if conversion fails.
        """
        return self._convert_and_export(
            source,
            export_format="text",
            headers=headers,
            raises_on_error=raises_on_error,
            max_num_pages=max_num_pages,
            max_file_size=max_file_size,
        )

    def convert_to_html(
        self,
        source: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        """Convert a local file or URL to HTML format using Docling.

        Args:
            source (str): Local file path or URL to the document.
            headers (Optional[Dict[str, str]]): Optional HTTP headers used for URL fetching.
            raises_on_error (bool): If True, raises conversion exceptions in Docling internals.
            max_num_pages (Optional[int]): Maximum number of pages to process from the source.
            max_file_size (Optional[int]): Maximum file size in bytes allowed for processing.

        Returns:
            str: The converted HTML content, or an error message if conversion fails.
        """
        return self._convert_and_export(
            source,
            export_format="html",
            headers=headers,
            raises_on_error=raises_on_error,
            max_num_pages=max_num_pages,
            max_file_size=max_file_size,
        )

    def convert_to_html_split_page(
        self,
        source: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        """Convert a local file or URL to split-page HTML format using Docling.

        Args:
            source (str): Local file path or URL to the document.
            headers (Optional[Dict[str, str]]): Optional HTTP headers used for URL fetching.
            raises_on_error (bool): If True, raises conversion exceptions in Docling internals.
            max_num_pages (Optional[int]): Maximum number of pages to process from the source.
            max_file_size (Optional[int]): Maximum file size in bytes allowed for processing.

        Returns:
            str: The converted split-page HTML content, or an error message if conversion fails.
        """
        return self._convert_and_export(
            source,
            export_format="html_split_page",
            headers=headers,
            raises_on_error=raises_on_error,
            max_num_pages=max_num_pages,
            max_file_size=max_file_size,
        )

    def convert_to_json(
        self,
        source: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        """Convert a local file or URL to JSON format using Docling.

        Args:
            source (str): Local file path or URL to the document.
            headers (Optional[Dict[str, str]]): Optional HTTP headers used for URL fetching.
            raises_on_error (bool): If True, raises conversion exceptions in Docling internals.
            max_num_pages (Optional[int]): Maximum number of pages to process from the source.
            max_file_size (Optional[int]): Maximum file size in bytes allowed for processing.

        Returns:
            str: The converted JSON content, or an error message if conversion fails.
        """
        return self._convert_and_export(
            source,
            export_format="json",
            headers=headers,
            raises_on_error=raises_on_error,
            max_num_pages=max_num_pages,
            max_file_size=max_file_size,
        )

    def convert_to_yaml(
        self,
        source: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        """Convert a local file or URL to YAML format using Docling.

        Args:
            source (str): Local file path or URL to the document.
            headers (Optional[Dict[str, str]]): Optional HTTP headers used for URL fetching.
            raises_on_error (bool): If True, raises conversion exceptions in Docling internals.
            max_num_pages (Optional[int]): Maximum number of pages to process from the source.
            max_file_size (Optional[int]): Maximum file size in bytes allowed for processing.

        Returns:
            str: The converted YAML content, or an error message if conversion fails.
        """
        return self._convert_and_export(
            source,
            export_format="yaml",
            headers=headers,
            raises_on_error=raises_on_error,
            max_num_pages=max_num_pages,
            max_file_size=max_file_size,
        )

    def convert_to_doctags(
        self,
        source: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        """Convert a local file or URL to DocTags format using Docling.

        Args:
            source (str): Local file path or URL to the document.
            headers (Optional[Dict[str, str]]): Optional HTTP headers used for URL fetching.
            raises_on_error (bool): If True, raises conversion exceptions in Docling internals.
            max_num_pages (Optional[int]): Maximum number of pages to process from the source.
            max_file_size (Optional[int]): Maximum file size in bytes allowed for processing.

        Returns:
            str: The converted DocTags content, or an error message if conversion fails.
        """
        return self._convert_and_export(
            source,
            export_format="doctags",
            headers=headers,
            raises_on_error=raises_on_error,
            max_num_pages=max_num_pages,
            max_file_size=max_file_size,
        )

    def convert_to_vtt(
        self,
        source: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        """Convert a local file or URL to VTT format using Docling.

        Args:
            source (str): Local file path or URL to the document.
            headers (Optional[Dict[str, str]]): Optional HTTP headers used for URL fetching.
            raises_on_error (bool): If True, raises conversion exceptions in Docling internals.
            max_num_pages (Optional[int]): Maximum number of pages to process from the source.
            max_file_size (Optional[int]): Maximum file size in bytes allowed for processing.

        Returns:
            str: The converted VTT content, or an error message if conversion fails.
        """
        return self._convert_and_export(
            source,
            export_format="vtt",
            headers=headers,
            raises_on_error=raises_on_error,
            max_num_pages=max_num_pages,
            max_file_size=max_file_size,
        )

    def convert_string_content(
        self,
        content: str,
        source_format: str = "markdown",
        output_format: str = "markdown",
        name: Optional[str] = None,
    ) -> str:
        """Convert raw markdown or HTML string content using Docling convert_string.

        Args:
            content (str): Raw source content to convert.
            source_format (str): Input content format, one of markdown, md, or html.
            output_format (str): Export format used after conversion.
            name (Optional[str]): Optional document name associated with this in-memory source.

        Returns:
            str: Converted content in the selected output format, or an error message if conversion fails.
        """
        if not content:
            return "Error: No content provided"

        try:
            input_format = self._resolve_string_input_format(source_format)
            result = self.converter.convert_string(content=content, format=input_format, name=name)
            exported_content = self._export_document(result.document, output_format)
            return self._truncate_content(exported_content, output_format)
        except Exception as e:
            log_error(f"Error converting string content: {e}")
            return f"Error converting string content: {e}"

    def list_supported_parsers(self) -> str:
        """List all Docling-supported input parsers and active converter parser restrictions.

        Returns:
            str: A JSON payload with supported_input_parsers and active_allowed_parsers fields.
        """
        all_supported = sorted([input_format.name.lower() for input_format in InputFormat])
        converter_allowed_formats = getattr(self.converter, "allowed_formats", None)
        if isinstance(converter_allowed_formats, list):
            active_formats = sorted([input_format.name.lower() for input_format in converter_allowed_formats])
        else:
            active_formats = all_supported
        payload = {
            "supported_input_parsers": all_supported,
            "active_allowed_parsers": active_formats,
        }
        return json.dumps(payload, indent=2)

    def _convert_and_export(
        self,
        source: str,
        export_format: str,
        headers: Optional[Dict[str, str]] = None,
        raises_on_error: bool = True,
        max_num_pages: Optional[int] = None,
        max_file_size: Optional[int] = None,
    ) -> str:
        if not source:
            return "Error: No source provided"

        try:
            log_debug(f"Converting document with Docling: {source}")
            convert_kwargs: Dict[str, Any] = {
                "headers": headers,
                "raises_on_error": raises_on_error,
            }
            if max_num_pages is not None:
                convert_kwargs["max_num_pages"] = max_num_pages
            if max_file_size is not None:
                convert_kwargs["max_file_size"] = max_file_size

            result = self.converter.convert(source, **convert_kwargs)
            content = self._export_document(result.document, export_format)

            return self._truncate_content(content, export_format)
        except Exception as e:
            log_error(f"Error converting document: {e}")
            return f"Error converting document: {e}"

    def _export_document(self, document: Any, export_format: str) -> str:
        if export_format == "markdown":
            return document.export_to_markdown()
        if export_format == "text":
            return document.export_to_text()
        if export_format == "html":
            return document.export_to_html()
        if export_format == "html_split_page":
            return document.export_to_html(split_page_view=True)
        if export_format == "json":
            return json.dumps(document.export_to_dict(), indent=2)
        if export_format == "yaml":
            import yaml

            return yaml.safe_dump(document.export_to_dict())
        if export_format == "doctags":
            return document.export_to_doctags()
        if export_format == "vtt":
            document_obj: Any = document
            export_to_vtt = getattr(document_obj, "export_to_vtt", None)
            if not callable(export_to_vtt):
                raise ValueError("VTT export is not supported by the installed docling version")
            vtt_content = export_to_vtt()
            return vtt_content if isinstance(vtt_content, str) else str(vtt_content)

        raise ValueError(f"Unsupported export format {export_format}")

    def _resolve_string_input_format(self, source_format: str) -> Any:
        source_format_value = source_format.lower().strip()
        if source_format_value in ["markdown", "md"]:
            return InputFormat.MD
        if source_format_value == "html":
            return InputFormat.HTML
        raise ValueError("source_format must be one of: markdown, md, html")

    def _build_converter(
        self,
        allowed_input_formats: Optional[List[str]],
        format_options: Optional[Dict[Any, Any]],
        pdf_pipeline_options: Optional[PdfPipelineOptions],
        pdf_enable_ocr: Optional[bool],
        pdf_ocr_engine: Optional[str],
        pdf_ocr_lang: Optional[List[str]],
        pdf_force_full_page_ocr: Optional[bool],
        pdf_enable_table_structure: Optional[bool],
        pdf_enable_picture_description: Optional[bool],
        pdf_enable_picture_classification: Optional[bool],
        pdf_document_timeout: Optional[float],
        pdf_enable_remote_services: Optional[bool],
    ) -> DocumentConverter:
        options = dict(format_options or {})
        resolved_allowed_formats = self._resolve_allowed_input_formats(allowed_input_formats)

        pdf_options = self._build_pdf_pipeline_options(
            pdf_pipeline_options=pdf_pipeline_options,
            pdf_enable_ocr=pdf_enable_ocr,
            pdf_ocr_engine=pdf_ocr_engine,
            pdf_ocr_lang=pdf_ocr_lang,
            pdf_force_full_page_ocr=pdf_force_full_page_ocr,
            pdf_enable_table_structure=pdf_enable_table_structure,
            pdf_enable_picture_description=pdf_enable_picture_description,
            pdf_enable_picture_classification=pdf_enable_picture_classification,
            pdf_document_timeout=pdf_document_timeout,
            pdf_enable_remote_services=pdf_enable_remote_services,
        )
        if pdf_options:
            options[InputFormat.PDF] = PdfFormatOption(pipeline_options=pdf_options)

        converter_kwargs: Dict[str, Any] = {}
        if resolved_allowed_formats is not None:
            converter_kwargs["allowed_formats"] = resolved_allowed_formats
        if options:
            converter_kwargs["format_options"] = options

        if converter_kwargs:
            return DocumentConverter(**converter_kwargs)
        return DocumentConverter()

    def _resolve_allowed_input_formats(self, allowed_input_formats: Optional[List[str]]) -> Optional[List[InputFormat]]:
        if allowed_input_formats is None:
            return None

        alias_map = {
            "markdown": "md",
        }
        resolved_formats: List[InputFormat] = []
        valid_names = sorted([input_format.name.lower() for input_format in InputFormat])

        for input_format_name in allowed_input_formats:
            normalized_name = input_format_name.lower().strip()

            if normalized_name == "xml":
                xml_variants = [name for name in valid_names if name.startswith("xml_")]
                variants_message = ", ".join(xml_variants) if xml_variants else "explicit xml_* parser"
                raise ValueError(f"Ambiguous input parser 'xml'. Use one of: {variants_message}.")

            normalized_name = alias_map.get(normalized_name, normalized_name)

            resolved = None
            for input_format in InputFormat:
                if input_format.name.lower() == normalized_name:
                    resolved = input_format
                    break

            if resolved is None:
                raise ValueError(
                    f"Invalid input parser '{input_format_name}'. Expected one of: {', '.join(valid_names)}"
                )

            resolved_formats.append(resolved)

        return resolved_formats

    def _build_pdf_pipeline_options(
        self,
        pdf_pipeline_options: Optional[PdfPipelineOptions],
        pdf_enable_ocr: Optional[bool],
        pdf_ocr_engine: Optional[str],
        pdf_ocr_lang: Optional[List[str]],
        pdf_force_full_page_ocr: Optional[bool],
        pdf_enable_table_structure: Optional[bool],
        pdf_enable_picture_description: Optional[bool],
        pdf_enable_picture_classification: Optional[bool],
        pdf_document_timeout: Optional[float],
        pdf_enable_remote_services: Optional[bool],
    ) -> Optional[PdfPipelineOptions]:
        if pdf_pipeline_options is not None:
            return pdf_pipeline_options

        if (
            pdf_enable_ocr is None
            and pdf_ocr_engine is None
            and pdf_ocr_lang is None
            and pdf_force_full_page_ocr is None
            and pdf_enable_table_structure is None
            and pdf_enable_picture_description is None
            and pdf_enable_picture_classification is None
            and pdf_document_timeout is None
            and pdf_enable_remote_services is None
        ):
            return None

        options = PdfPipelineOptions()
        if pdf_enable_ocr is not None:
            options.do_ocr = pdf_enable_ocr
        if pdf_enable_table_structure is not None:
            options.do_table_structure = pdf_enable_table_structure
        if pdf_enable_picture_description is not None:
            options.do_picture_description = pdf_enable_picture_description
        if pdf_enable_picture_classification is not None:
            options.do_picture_classification = pdf_enable_picture_classification
        if pdf_document_timeout is not None:
            options.document_timeout = pdf_document_timeout
        if pdf_enable_remote_services is not None:
            options.enable_remote_services = pdf_enable_remote_services

        ocr_options = self._build_ocr_options(
            engine=pdf_ocr_engine,
            lang=pdf_ocr_lang,
            force_full_page_ocr=pdf_force_full_page_ocr,
        )
        if ocr_options is not None:
            options.ocr_options = ocr_options
            if pdf_enable_ocr is None:
                options.do_ocr = True

        return options

    def _build_ocr_options(
        self,
        engine: Optional[str],
        lang: Optional[List[str]],
        force_full_page_ocr: Optional[bool],
    ) -> Optional[Any]:
        if not engine:
            return None

        engine_value = engine.lower()
        languages = lang or []
        kwargs: Dict[str, Any] = {"lang": languages}
        if force_full_page_ocr is not None:
            kwargs["force_full_page_ocr"] = force_full_page_ocr

        engine_map = {
            "auto": OcrAutoOptions,
            "easyocr": EasyOcrOptions,
            "tesseract": TesseractOcrOptions,
            "tesseract_cli": TesseractCliOcrOptions,
            "ocrmac": OcrMacOptions,
            "rapidocr": RapidOcrOptions,
        }

        ocr_cls = engine_map.get(engine_value)
        if ocr_cls is not None:
            dependency_hints = {
                "easyocr": "Install optional dependencies, e.g. `uv pip install easyocr`.",
                "tesseract": "Install the Python tesseract dependencies and ensure tesseract OCR is available.",
                "tesseract_cli": "Install the tesseract CLI binary and ensure it is available on PATH.",
                "rapidocr": "Install optional dependencies, e.g. `uv pip install rapidocr_onnxruntime`.",
                "ocrmac": "OCRMac is only available on macOS with required native dependencies.",
            }
            try:
                return ocr_cls(**kwargs)
            except Exception as e:
                hint = dependency_hints.get(engine_value, "Install required OCR runtime dependencies for this engine.")
                raise RuntimeError(f"Failed to initialize OCR engine '{engine}'. {hint}") from e

        valid_engines = list(engine_map.keys())
        raise ValueError(f"Invalid OCR engine '{engine}'. Expected one of: {', '.join(valid_engines)}.")

    def _truncate_content(self, content: str, export_format: str) -> str:
        if self.max_chars and len(content) > self.max_chars:
            if export_format == "json":
                # Keep JSON output valid even when truncating long content.
                truncated_payload = {
                    "truncated": True,
                    "max_chars": self.max_chars,
                    "content": content[: self.max_chars] + "...",
                }
                return json.dumps(truncated_payload, indent=2)
            if export_format == "yaml":
                import yaml

                truncated_payload = {
                    "truncated": True,
                    "max_chars": self.max_chars,
                    "content": content[: self.max_chars] + "...",
                }
                return yaml.safe_dump(truncated_payload)
            return content[: self.max_chars] + "..."
        return content



"""Prompt/message building and deep-copy helpers for Team."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from agno.team.team import Team

import json
from collections import ChainMap
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)

from pydantic import BaseModel

from agno.media import Audio, File, Image, Video
from agno.models.base import Model
from agno.models.message import Message, MessageReferences
from agno.models.response import ModelResponse
from agno.run import RunContext
from agno.run.messages import RunMessages
from agno.run.team import (
    TeamRunOutput,
)
from agno.session import TeamSession
from agno.tools import Toolkit
from agno.tools.function import Function
from agno.utils.agent import (
    aexecute_instructions,
    aexecute_system_message,
    execute_instructions,
    execute_system_message,
)
from agno.utils.common import is_typed_dict
from agno.utils.log import (
    log_debug,
    log_warning,
)
from agno.utils.message import filter_tool_calls, get_text_from_message
from agno.utils.team import (
    get_member_id,
)
from agno.utils.timer import Timer


def _get_tool_names(member: Any, async_mode: bool = False) -> List[str]:
    """Extract tool names from a member's tools list."""
    tool_names: List[str] = []
    if member.tools is None or not isinstance(member.tools, list):
        return tool_names
    for _tool in member.tools:
        if isinstance(_tool, Toolkit):
            toolkit_functions = _tool.get_async_functions() if async_mode else _tool.get_functions()
            for _func in toolkit_functions.values():
                if _func.entrypoint:
                    tool_names.append(_func.name)
        elif isinstance(_tool, Function) and _tool.entrypoint:
            tool_names.append(_tool.name)
        elif callable(_tool):
            tool_names.append(_tool.__name__)
        elif isinstance(_tool, dict) and "name" in _tool and _tool.get("name") is not None:
            tool_names.append(_tool["name"])
        else:
            tool_names.append(str(_tool))
    return tool_names


def get_members_system_message_content(
    team: "Team", indent: int = 0, run_context: Optional["RunContext"] = None, async_mode: bool = False
) -> str:
    from agno.team.team import Team
    from agno.utils.callables import get_resolved_members

    pad = " " * indent
    content = ""
    resolved_members = get_resolved_members(team, run_context)
    if resolved_members is None or len(resolved_members) == 0:
        return content
    for member in resolved_members:
        member_id = get_member_id(member)

        if isinstance(member, Team):
            content += f'{pad}<member id="{member_id}" name="{member.name}" type="team">\n'
            if member.description is not None:
                content += f"{pad}  Description: {member.description}\n"
            if member.members is not None:
                content += member.get_members_system_message_content(
                    indent=indent + 2, run_context=run_context, async_mode=async_mode
                )
            content += f"{pad}</member>\n"
        else:
            content += f'{pad}<member id="{member_id}" name="{member.name}">\n'
            if member.role is not None:
                content += f"{pad}  Role: {member.role}\n"
            if member.description is not None:
                content += f"{pad}  Description: {member.description}\n"
            if team.add_member_tools_to_context:
                tool_names = _get_tool_names(member, async_mode=async_mode)
                if tool_names:
                    content += f"{pad}  Tools: {', '.join(tool_names)}\n"
            content += f"{pad}</member>\n"

    return content


def _get_opening_prompt() -> str:
    """Opening identity statement for the team leader."""
    return (
        "You coordinate a team of specialized AI agents to fulfill the user's request. "
        "Delegate to members when their expertise or tools are needed. "
        "For straightforward requests you can handle directly — including using your own tools — respond without delegating.\n"
    )


def _get_mode_instructions(team: "Team") -> str:
    """Return the mode-specific <how_to_respond> block."""
    from agno.team.mode import TeamMode

    content = "\n<how_to_respond>\n"

    if team.mode == TeamMode.tasks:
        content += (
            "You operate in autonomous task mode. Decompose the user's goal into discrete tasks, "
            "execute them by delegating to team members, and deliver the final result.\n\n"
            "Planning:\n"
            "- Break the goal into tasks with clear, actionable titles and self-contained descriptions. "
            "Each task should be a single unit of work for one member.\n"
            "- Assign each task to the member whose role and tools are best suited.\n"
            "- Set `depends_on` when a task requires another task's output. "
            "Leave tasks independent when they can run in any order.\n\n"
            "Execution:\n"
            "- Use `execute_task` for sequential or dependent tasks.\n"
            "- Use `execute_tasks_parallel` for groups of independent tasks to maximize throughput.\n"
            "- Review each result before proceeding. If a task fails, decide whether to retry with the same member, "
            "reassign to a different member, or adjust the plan.\n\n"
            "Completion:\n"
            "- When all tasks are done and results are satisfactory, call `mark_all_complete` with a summary of the outcome.\n"
            "- Use `list_tasks` to check progress at any point, and `add_task_note` to record observations.\n\n"
            "Write task descriptions that give the member everything they need: "
            "the objective, relevant context from the conversation or prior task results, and what a good result looks like.\n"
        )
    elif team.mode == TeamMode.route:
        content += (
            "You operate in route mode. For requests that need member expertise, "
            "identify the single best member and delegate to them — their response is returned directly to the user. "
            "For requests you can handle directly — simple questions, using your own tools, or general conversation — "
            "respond without delegating.\n\n"
            "When routing to a member:\n"
            "- Analyze the request to determine which member's role and tools are the best match.\n"
            "- Delegate to exactly one member. Use only the member's ID — do not prefix it with the team ID.\n"
            "- Write the task to faithfully represent the user's full intent. Do not reinterpret or narrow the request.\n"
            "- If no member is a clear fit, choose the closest match and include any additional context the member might need.\n"
        )
    elif team.mode == TeamMode.broadcast:
        content += (
            "You operate in broadcast mode. For requests that benefit from multiple perspectives, "
            "send the request to all members simultaneously and synthesize their collective responses. "
            "For requests you can handle directly — simple questions, using your own tools, or general conversation — "
            "respond without delegating.\n\n"
            "When broadcasting:\n"
            "- Call `delegate_task_to_members` exactly once with a clear task description. "
            "This sends the task to every member in parallel.\n"
            "- Write the task so each member can respond independently from their own perspective.\n\n"
            "After receiving member responses:\n"
            "- Compare perspectives: note agreements, highlight complementary insights, and reconcile any contradictions.\n"
            "- Synthesize into a unified answer that integrates the strongest contributions thematically — "
            "do not list each member's response sequentially.\n"
        )
    else:
        # coordinate mode (default)
        content += (
            "You operate in coordinate mode. For requests that need member expertise, "
            "select the best member(s), delegate with clear task descriptions, and synthesize their outputs "
            "into a unified response. For requests you can handle directly — simple questions, "
            "using your own tools, or general conversation — respond without delegating.\n\n"
            "Delegation:\n"
            "- Match each sub-task to the member whose role and tools are the best fit. "
            "Delegate to multiple members when the request spans different areas of expertise.\n"
            "- Write task descriptions that are self-contained: state the goal, provide relevant context "
            "from the conversation, and describe what a good result looks like.\n"
            "- Use only the member's ID when delegating — do not prefix it with the team ID.\n\n"
            "After receiving member responses:\n"
            "- If a response is incomplete or off-target, re-delegate with clearer instructions or try a different member.\n"
            "- Synthesize all results into a single coherent response. Resolve contradictions, fill gaps with your own "
            "reasoning, and add structure — do not simply concatenate member outputs.\n"
        )

    content += "</how_to_respond>\n\n"
    return content


def _build_team_context(
    team: "Team",
    run_context: Optional["RunContext"] = None,
    async_mode: bool = False,
) -> str:
    """Build the opening + team_members + how_to_respond blocks.

    Shared between sync and async system-message builders.
    """
    from agno.utils.callables import get_resolved_members

    content = ""
    resolved_members = get_resolved_members(team, run_context)
    if resolved_members is not None and len(resolved_members) > 0:
        content += _get_opening_prompt()
        content += "\n<team_members>\n"
        content += team.get_members_system_message_content(run_context=run_context, async_mode=async_mode)
        if team.get_member_information_tool:
            content += "If you need to get information about your team members, you can use the `get_member_information` tool at any time.\n"
        content += "</team_members>\n"
        content += _get_mode_instructions(team)
    return content


def _build_identity_sections(
    team: "Team",
    instructions: List[str],
) -> str:
    """Build description, role, and instructions sections.

    Shared between sync and async system-message builders.
    """
    content = ""
    if team.description is not None:
        content += f"<description>\n{team.description}\n</description>\n\n"

    if team.role is not None:
        content += f"<your_role>\n{team.role}\n</your_role>\n\n"

    if len(instructions) > 0:
        if team.use_instruction_tags:
            content += "<instructions>"
            if len(instructions) > 1:
                for _upi in instructions:
                    content += f"\n- {_upi}"
            else:
                content += "\n" + instructions[0]
            content += "\n</instructions>\n\n"
        else:
            if len(instructions) > 1:
                for _upi in instructions:
                    content += f"- {_upi}\n"
            else:
                content += instructions[0] + "\n\n"
    return content


def _build_trailing_sections(
    team: "Team",
    *,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    additional_information: List[str],
    tools: Optional[List[Union[Function, dict]]] = None,
    output_schema: Optional[Any] = None,
    run_context: Optional[RunContext] = None,
    session_state: Optional[Dict[str, Any]] = None,
    add_session_state_to_context: Optional[bool] = None,
) -> str:
    """Build media, additional info, tool instructions, and other trailing sections.

    Shared between sync and async system-message builders.
    """
    content = ""

    # Attached media
    if audio is not None or images is not None or videos is not None or files is not None:
        content += "<attached_media>\n"
        content += "You have the following media attached to your message:\n"
        if audio is not None and len(audio) > 0:
            content += " - Audio\n"
        if images is not None and len(images) > 0:
            content += " - Images\n"
        if videos is not None and len(videos) > 0:
            content += " - Videos\n"
        if files is not None and len(files) > 0:
            content += " - Files\n"
        content += "</attached_media>\n\n"

    # Additional information
    if len(additional_information) > 0:
        content += "<additional_information>"
        for _ai in additional_information:
            content += f"\n- {_ai}"
        content += "\n</additional_information>\n\n"

    # Tool instructions
    if team._tool_instructions is not None:
        for _ti in team._tool_instructions:
            content += f"{_ti}\n"

    system_message_from_model = team.model.get_system_message_for_model(tools)  # type: ignore[union-attr]
    if system_message_from_model is not None:
        content += system_message_from_model

    if team.expected_output is not None:
        content += f"<expected_output>\n{team.expected_output.strip()}\n</expected_output>\n\n"

    if team.additional_context is not None:
        content += f"<additional_context>\n{team.additional_context.strip()}\n</additional_context>\n\n"

    if add_session_state_to_context and session_state is not None:
        content += _get_formatted_session_state_for_system_message(team, session_state)

    # JSON output prompt
    if (
        output_schema is not None
        and team.parser_model is None
        and team.model
        and not (
            (team.model.supports_native_structured_outputs or team.model.supports_json_schema_outputs)
            and not team.use_json_mode
        )
    ):
        content += f"{_get_json_output_prompt(team, output_schema)}"

    return content


def get_system_message(
    team: "Team",
    session: TeamSession,
    run_context: Optional[RunContext] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    tools: Optional[List[Union[Function, dict]]] = None,
    add_session_state_to_context: Optional[bool] = None,
) -> Optional[Message]:
    """Get the system message for the team.

    1. If the system_message is provided, use that.
    2. If build_context is False, return None.
    3. Build and return the default system message for the Team.
    """

    # Extract values from run_context
    from agno.team._init import _has_async_db, _set_memory_manager

    session_state = run_context.session_state if run_context else None
    user_id = run_context.user_id if run_context else None

    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    # 1. If the system_message is provided, use that.
    if team.system_message is not None:
        if isinstance(team.system_message, Message):
            return team.system_message

        sys_message_content: str = ""
        if isinstance(team.system_message, str):
            sys_message_content = team.system_message
        elif callable(team.system_message):
            sys_message_content = execute_system_message(
                system_message=team.system_message,
                agent=cast(Any, team),
                team=cast(Any, team),
                session_state=session_state,
                run_context=run_context,
            )
            if not isinstance(sys_message_content, str):
                raise Exception("system_message must return a string")

        # Format the system message with the session state variables
        if team.resolve_in_context:
            sys_message_content = _format_message_with_state_variables(
                team,
                sys_message_content,
                run_context=run_context,
            )

        # type: ignore
        return Message(role=team.system_message_role, content=sys_message_content)

    # 1. Build and return the default system message for the Team.
    # 1.1 Build the list of instructions for the system message
    team.model = cast(Model, team.model)
    instructions: List[str] = []
    if team.instructions is not None:
        _instructions = team.instructions
        if callable(team.instructions):
            _instructions = execute_instructions(
                instructions=team.instructions,
                agent=cast(Any, team),
                team=cast(Any, team),
                session_state=session_state,
                run_context=run_context,
            )

        if isinstance(_instructions, str):
            instructions.append(_instructions)
        elif isinstance(_instructions, list):
            instructions.extend(_instructions)

    # 1.2 Add instructions from the Model
    _model_instructions = team.model.get_instructions_for_model(tools)
    if _model_instructions is not None:
        instructions.extend(_model_instructions)

    # 1.3 Build a list of additional information for the system message
    additional_information: List[str] = []
    # 1.3.1 Add instructions for using markdown
    if team.markdown and output_schema is None:
        additional_information.append("Use markdown to format your answers.")
    # 1.3.2 Add the current datetime
    if team.add_datetime_to_context:
        from datetime import datetime

        tz = None

        if team.timezone_identifier:
            try:
                from zoneinfo import ZoneInfo

                tz = ZoneInfo(team.timezone_identifier)
            except Exception:
                log_warning("Invalid timezone identifier")

        time = datetime.now(tz) if tz else datetime.now()

        if team.datetime_format:
            formatted_time = time.strftime(team.datetime_format)
        else:
            formatted_time = str(time)

        additional_information.append(f"The current time is {formatted_time}.")

    # 1.3.3 Add the current location
    if team.add_location_to_context:
        from agno.utils.location import get_location

        location = get_location()
        if location:
            location_str = ", ".join(
                filter(None, [location.get("city"), location.get("region"), location.get("country")])
            )
            if location_str:
                additional_information.append(f"Your approximate location is: {location_str}.")

    # 1.3.4 Add team name if provided
    if team.name is not None and team.add_name_to_context:
        additional_information.append(f"Your name is: {team.name}.")

    # 2 Build the default system message for the Team.
    system_message_content: str = ""

    # 2.1 Opening + team members + mode instructions
    system_message_content += _build_team_context(team, run_context=run_context)

    # 2.2 Identity sections: description, role, instructions
    system_message_content += _build_identity_sections(team, instructions)

    # 2.3 Knowledge base instructions
    if team.knowledge is not None and team.search_knowledge and team.add_search_knowledge_instructions:
        build_context_fn = getattr(team.knowledge, "build_context", None)
        if callable(build_context_fn):
            knowledge_context = build_context_fn(
                enable_agentic_filters=team.enable_agentic_knowledge_filters,
            )
            if knowledge_context:
                system_message_content += knowledge_context + "\n"

    # 2.4 Memories
    if team.add_memories_to_context:
        _memory_manager_not_set = False
        if not user_id:
            user_id = "default"
        if team.memory_manager is None:
            _set_memory_manager(team)
            _memory_manager_not_set = True
        if _has_async_db(team):
            raise ValueError(
                "Sync get_system_message cannot retrieve user memories with an async database. "
                "Use aget_system_message instead."
            )
        user_memories = team.memory_manager.get_user_memories(user_id=user_id)  # type: ignore
        if user_memories and len(user_memories) > 0:
            system_message_content += "You have access to user info and preferences from previous interactions that you can use to personalize your response:\n\n"
            system_message_content += "<memories_from_previous_interactions>"
            for _memory in user_memories:  # type: ignore
                system_message_content += f"\n- {_memory.memory}"
            system_message_content += "\n</memories_from_previous_interactions>\n\n"
            system_message_content += (
                "Note: this information is from previous interactions and may be updated in this conversation. "
                "You should always prefer information from this conversation over the past memories.\n"
            )
        else:
            system_message_content += (
                "You have the capability to retain memories from previous interactions with the user, "
                "but have not had any interactions with the user yet.\n"
            )
        if _memory_manager_not_set:
            team.memory_manager = None

        if team.enable_agentic_memory:
            system_message_content += (
                "\n<updating_user_memories>\n"
                "- You have access to the `update_user_memory` tool that you can use to add new memories, update existing memories, delete memories, or clear all memories.\n"
                "- If the user's message includes information that should be captured as a memory, use the `update_user_memory` tool to update your memory database.\n"
                "- Memories should include details that could personalize ongoing interactions with the user.\n"
                "- Use this tool to add new memories or update existing memories that you identify in the conversation.\n"
                "- Use this tool if the user asks to update their memory, delete a memory, or clear all memories.\n"
                "- If you use the `update_user_memory` tool, remember to pass on the response to the user.\n"
                "</updating_user_memories>\n\n"
            )

    # 2.5 Session summary
    if team.add_session_summary_to_context and session.summary is not None:
        system_message_content += "Here is a brief summary of your previous interactions:\n\n"
        system_message_content += "<summary_of_previous_interactions>\n"
        system_message_content += session.summary.summary
        system_message_content += "\n</summary_of_previous_interactions>\n\n"
        system_message_content += (
            "Note: this information is from previous interactions and may be outdated. "
            "You should ALWAYS prefer information from this conversation over the past summary.\n\n"
        )

    # 2.6 Trailing sections: media, additional info, tools, expected output, etc.
    system_message_content += _build_trailing_sections(
        team,
        audio=audio,
        images=images,
        videos=videos,
        files=files,
        additional_information=additional_information,
        tools=tools,
        output_schema=output_schema,
        run_context=run_context,
        session_state=session_state,
        add_session_state_to_context=add_session_state_to_context,
    )

    # Format the full system message with dependencies and session state variables
    if team.resolve_in_context:
        system_message_content = _format_message_with_state_variables(
            team,
            system_message_content,
            run_context=run_context,
        )

    return Message(role=team.system_message_role, content=system_message_content.strip())


async def aget_system_message(
    team: "Team",
    session: TeamSession,
    run_context: Optional[RunContext] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    tools: Optional[List[Union[Function, dict]]] = None,
    add_session_state_to_context: Optional[bool] = None,
) -> Optional[Message]:
    """Get the system message for the team."""

    # Extract values from run_context
    from agno.team._init import _has_async_db, _set_memory_manager

    session_state = run_context.session_state if run_context else None
    user_id = run_context.user_id if run_context else None

    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    # 1. If the system_message is provided, use that.
    if team.system_message is not None:
        if isinstance(team.system_message, Message):
            return team.system_message

        sys_message_content: str = ""
        if isinstance(team.system_message, str):
            sys_message_content = team.system_message
        elif callable(team.system_message):
            sys_message_content = await aexecute_system_message(
                system_message=team.system_message,
                agent=cast(Any, team),
                team=cast(Any, team),
                session_state=session_state,
                run_context=run_context,
            )
            if not isinstance(sys_message_content, str):
                raise Exception("system_message must return a string")

        # Format the system message with the session state variables
        if team.resolve_in_context:
            sys_message_content = _format_message_with_state_variables(
                team,
                sys_message_content,
                run_context=run_context,
            )

        # type: ignore
        return Message(role=team.system_message_role, content=sys_message_content)

    # 1. Build and return the default system message for the Team.
    # 1.1 Build the list of instructions for the system message
    team.model = cast(Model, team.model)
    instructions: List[str] = []
    if team.instructions is not None:
        _instructions = team.instructions
        if callable(team.instructions):
            _instructions = await aexecute_instructions(
                instructions=team.instructions,
                agent=cast(Any, team),
                team=cast(Any, team),
                session_state=session_state,
                run_context=run_context,
            )

        if isinstance(_instructions, str):
            instructions.append(_instructions)
        elif isinstance(_instructions, list):
            instructions.extend(_instructions)

    # 1.2 Add instructions from the Model
    _model_instructions = team.model.get_instructions_for_model(tools)
    if _model_instructions is not None:
        instructions.extend(_model_instructions)

    # 1.3 Build a list of additional information for the system message
    additional_information: List[str] = []
    # 1.3.1 Add instructions for using markdown
    if team.markdown and output_schema is None:
        additional_information.append("Use markdown to format your answers.")
    # 1.3.2 Add the current datetime
    if team.add_datetime_to_context:
        from datetime import datetime

        tz = None

        if team.timezone_identifier:
            try:
                from zoneinfo import ZoneInfo

                tz = ZoneInfo(team.timezone_identifier)
            except Exception:
                log_warning("Invalid timezone identifier")

        time = datetime.now(tz) if tz else datetime.now()

        if team.datetime_format:
            formatted_time = time.strftime(team.datetime_format)
        else:
            formatted_time = str(time)

        additional_information.append(f"The current time is {formatted_time}.")

    # 1.3.3 Add the current location
    if team.add_location_to_context:
        from agno.utils.location import get_location

        location = get_location()
        if location:
            location_str = ", ".join(
                filter(None, [location.get("city"), location.get("region"), location.get("country")])
            )
            if location_str:
                additional_information.append(f"Your approximate location is: {location_str}.")

    # 1.3.4 Add team name if provided
    if team.name is not None and team.add_name_to_context:
        additional_information.append(f"Your name is: {team.name}.")

    # 2 Build the default system message for the Team.
    system_message_content: str = ""

    # 2.1 Opening + team members + mode instructions
    system_message_content += _build_team_context(team, run_context=run_context, async_mode=True)

    # 2.2 Identity sections: description, role, instructions
    system_message_content += _build_identity_sections(team, instructions)

    # 2.3 Knowledge base instructions
    if team.knowledge is not None and team.search_knowledge and team.add_search_knowledge_instructions:
        build_context_fn = getattr(team.knowledge, "build_context", None)
        if callable(build_context_fn):
            knowledge_context = build_context_fn(
                enable_agentic_filters=team.enable_agentic_knowledge_filters,
            )
            if knowledge_context:
                system_message_content += knowledge_context + "\n"

    # 2.4 Memories
    if team.add_memories_to_context:
        _memory_manager_not_set = False
        if not user_id:
            user_id = "default"
        if team.memory_manager is None:
            _set_memory_manager(team)
            _memory_manager_not_set = True

        if _has_async_db(team):
            user_memories = await team.memory_manager.aget_user_memories(user_id=user_id)  # type: ignore
        else:
            user_memories = team.memory_manager.get_user_memories(user_id=user_id)  # type: ignore

        if user_memories and len(user_memories) > 0:
            system_message_content += "You have access to user info and preferences from previous interactions that you can use to personalize your response:\n\n"
            system_message_content += "<memories_from_previous_interactions>"
            for _memory in user_memories:  # type: ignore
                system_message_content += f"\n- {_memory.memory}"
            system_message_content += "\n</memories_from_previous_interactions>\n\n"
            system_message_content += (
                "Note: this information is from previous interactions and may be updated in this conversation. "
                "You should always prefer information from this conversation over the past memories.\n"
            )
        else:
            system_message_content += (
                "You have the capability to retain memories from previous interactions with the user, "
                "but have not had any interactions with the user yet.\n"
            )
        if _memory_manager_not_set:
            team.memory_manager = None

        if team.enable_agentic_memory:
            system_message_content += (
                "\n<updating_user_memories>\n"
                "- You have access to the `update_user_memory` tool that you can use to add new memories, update existing memories, delete memories, or clear all memories.\n"
                "- If the user's message includes information that should be captured as a memory, use the `update_user_memory` tool to update your memory database.\n"
                "- Memories should include details that could personalize ongoing interactions with the user.\n"
                "- Use this tool to add new memories or update existing memories that you identify in the conversation.\n"
                "- Use this tool if the user asks to update their memory, delete a memory, or clear all memories.\n"
                "- If you use the `update_user_memory` tool, remember to pass on the response to the user.\n"
                "</updating_user_memories>\n\n"
            )

    # 2.5 Session summary
    if team.add_session_summary_to_context and session.summary is not None:
        system_message_content += "Here is a brief summary of your previous interactions:\n\n"
        system_message_content += "<summary_of_previous_interactions>\n"
        system_message_content += session.summary.summary
        system_message_content += "\n</summary_of_previous_interactions>\n\n"
        system_message_content += (
            "Note: this information is from previous interactions and may be outdated. "
            "You should ALWAYS prefer information from this conversation over the past summary.\n\n"
        )

    # 2.6 Trailing sections: media, additional info, tools, expected output, etc.
    system_message_content += _build_trailing_sections(
        team,
        audio=audio,
        images=images,
        videos=videos,
        files=files,
        additional_information=additional_information,
        tools=tools,
        output_schema=output_schema,
        run_context=run_context,
        session_state=session_state,
        add_session_state_to_context=add_session_state_to_context,
    )

    # Format the full system message with dependencies and session state variables
    if team.resolve_in_context:
        system_message_content = _format_message_with_state_variables(
            team,
            system_message_content,
            run_context=run_context,
        )

    return Message(role=team.system_message_role, content=system_message_content.strip())


def _get_formatted_session_state_for_system_message(team: "Team", session_state: Dict[str, Any]) -> str:
    return f"\n<session_state>\n{session_state}\n</session_state>\n\n"


def _get_run_messages(
    team: "Team",
    *,
    run_response: TeamRunOutput,
    run_context: RunContext,
    session: TeamSession,
    user_id: Optional[str] = None,
    input_message: Optional[Union[str, List, Dict, Message, BaseModel, List[Message]]] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    add_history_to_context: Optional[bool] = None,
    add_dependencies_to_context: Optional[bool] = None,
    add_session_state_to_context: Optional[bool] = None,
    tools: Optional[List[Union[Function, dict]]] = None,
    **kwargs: Any,
) -> RunMessages:
    """This function returns a RunMessages object with the following attributes:
        - system_message: The system message for this run
        - user_message: The user message for this run
        - messages: List of messages to send to the model

    To build the RunMessages object:
    1. Add system message to run_messages
    2. Add extra messages to run_messages
    3. Add history to run_messages
    4. Add messages to run_messages if provided (messages parameter first)
    5. Add user message to run_messages (message parameter second)

    """
    # Initialize the RunMessages object
    run_messages = RunMessages()

    # 1. Add system message to run_messages
    system_message = team.get_system_message(
        session=session,
        run_context=run_context,
        images=images,
        audio=audio,
        videos=videos,
        files=files,
        add_session_state_to_context=add_session_state_to_context,
        tools=tools,
    )
    if system_message is not None:
        run_messages.system_message = system_message
        run_messages.messages.append(system_message)

    # 2. Add extra messages to run_messages if provided
    if team.additional_input is not None:
        messages_to_add_to_run_response: List[Message] = []
        if run_messages.extra_messages is None:
            run_messages.extra_messages = []

        for _m in team.additional_input:
            if isinstance(_m, Message):
                messages_to_add_to_run_response.append(_m)
                run_messages.messages.append(_m)
                run_messages.extra_messages.append(_m)
            elif isinstance(_m, dict):
                try:
                    _m_parsed = Message.model_validate(_m)
                    messages_to_add_to_run_response.append(_m_parsed)
                    run_messages.messages.append(_m_parsed)
                    run_messages.extra_messages.append(_m_parsed)
                except Exception as e:
                    log_warning(f"Failed to validate message: {e}")
        # Add the extra messages to the run_response
        if len(messages_to_add_to_run_response) > 0:
            log_debug(f"Adding {len(messages_to_add_to_run_response)} extra messages")
            if run_response.additional_input is None:
                run_response.additional_input = messages_to_add_to_run_response
            else:
                run_response.additional_input.extend(messages_to_add_to_run_response)

    # 3. Add history to run_messages
    if add_history_to_context:
        from copy import deepcopy

        # Only skip messages from history when system_message_role is NOT a standard conversation role.
        # Standard conversation roles ("user", "assistant", "tool") should never be filtered
        # to preserve conversation continuity.
        skip_role = team.system_message_role if team.system_message_role not in ["user", "assistant", "tool"] else None

        history = session.get_messages(
            last_n_runs=team.num_history_runs,
            limit=team.num_history_messages,
            skip_roles=[skip_role] if skip_role else None,
            team_id=team.id if team.parent_team_id is not None else None,
        )

        if len(history) > 0:
            # Create a deep copy of the history messages to avoid modifying the original messages
            history_copy = [deepcopy(msg) for msg in history]

            # Tag each message as coming from history
            for _msg in history_copy:
                _msg.from_history = True

            # Filter tool calls from history messages
            if team.max_tool_calls_from_history is not None:
                filter_tool_calls(history_copy, team.max_tool_calls_from_history)

            log_debug(f"Adding {len(history_copy)} messages from history")

            # Extend the messages with the history
            run_messages.messages += history_copy

    # 5. Add user message to run_messages (message second as per Dirk's requirement)
    # 5.1 Build user message if message is None, str or list
    user_message = _get_user_message(
        team,
        run_response=run_response,
        run_context=run_context,
        input_message=input_message,
        user_id=user_id,
        audio=audio,
        images=images,
        videos=videos,
        files=files,
        add_dependencies_to_context=add_dependencies_to_context,
        **kwargs,
    )
    # Add user message to run_messages
    if user_message is not None:
        run_messages.user_message = user_message
        run_messages.messages.append(user_message)

    # Set messages on run_context so tool hooks can access the current message history
    run_context.messages = run_messages.messages

    return run_messages


async def _aget_run_messages(
    team: "Team",
    *,
    run_response: TeamRunOutput,
    run_context: RunContext,
    session: TeamSession,
    user_id: Optional[str] = None,
    input_message: Optional[Union[str, List, Dict, Message, BaseModel, List[Message]]] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    add_history_to_context: Optional[bool] = None,
    add_dependencies_to_context: Optional[bool] = None,
    add_session_state_to_context: Optional[bool] = None,
    tools: Optional[List[Union[Function, dict]]] = None,
    **kwargs: Any,
) -> RunMessages:
    """This function returns a RunMessages object with the following attributes:
        - system_message: The system message for this run
        - user_message: The user message for this run
        - messages: List of messages to send to the model

    To build the RunMessages object:
    1. Add system message to run_messages
    2. Add extra messages to run_messages
    3. Add history to run_messages
    4. Add messages to run_messages if provided (messages parameter first)
    5. Add user message to run_messages (message parameter second)

    """
    # Initialize the RunMessages object
    run_messages = RunMessages()

    # 1. Add system message to run_messages
    system_message = await team.aget_system_message(
        session=session,
        run_context=run_context,
        images=images,
        audio=audio,
        videos=videos,
        files=files,
        add_session_state_to_context=add_session_state_to_context,
        tools=tools,
    )
    if system_message is not None:
        run_messages.system_message = system_message
        run_messages.messages.append(system_message)

    # 2. Add extra messages to run_messages if provided
    if team.additional_input is not None:
        messages_to_add_to_run_response: List[Message] = []
        if run_messages.extra_messages is None:
            run_messages.extra_messages = []

        for _m in team.additional_input:
            if isinstance(_m, Message):
                messages_to_add_to_run_response.append(_m)
                run_messages.messages.append(_m)
                run_messages.extra_messages.append(_m)
            elif isinstance(_m, dict):
                try:
                    _m_parsed = Message.model_validate(_m)
                    messages_to_add_to_run_response.append(_m_parsed)
                    run_messages.messages.append(_m_parsed)
                    run_messages.extra_messages.append(_m_parsed)
                except Exception as e:
                    log_warning(f"Failed to validate message: {e}")
        # Add the extra messages to the run_response
        if len(messages_to_add_to_run_response) > 0:
            log_debug(f"Adding {len(messages_to_add_to_run_response)} extra messages")
            if run_response.additional_input is None:
                run_response.additional_input = messages_to_add_to_run_response
            else:
                run_response.additional_input.extend(messages_to_add_to_run_response)

    # 3. Add history to run_messages
    if add_history_to_context:
        from copy import deepcopy

        # Only skip messages from history when system_message_role is NOT a standard conversation role.
        # Standard conversation roles ("user", "assistant", "tool") should never be filtered
        # to preserve conversation continuity.
        skip_role = team.system_message_role if team.system_message_role not in ["user", "assistant", "tool"] else None
        history = session.get_messages(
            last_n_runs=team.num_history_runs,
            limit=team.num_history_messages,
            skip_roles=[skip_role] if skip_role else None,
            team_id=team.id if team.parent_team_id is not None else None,
        )

        if len(history) > 0:
            # Create a deep copy of the history messages to avoid modifying the original messages
            history_copy = [deepcopy(msg) for msg in history]

            # Tag each message as coming from history
            for _msg in history_copy:
                _msg.from_history = True

            # Filter tool calls from history messages
            if team.max_tool_calls_from_history is not None:
                filter_tool_calls(history_copy, team.max_tool_calls_from_history)

            log_debug(f"Adding {len(history_copy)} messages from history")

            # Extend the messages with the history
            run_messages.messages += history_copy

    # 5. Add user message to run_messages (message second as per Dirk's requirement)
    # 5.1 Build user message if message is None, str or list
    user_message = await _aget_user_message(
        team,
        run_response=run_response,
        run_context=run_context,
        input_message=input_message,
        user_id=user_id,
        audio=audio,
        images=images,
        videos=videos,
        files=files,
        add_dependencies_to_context=add_dependencies_to_context,
        **kwargs,
    )
    # Add user message to run_messages
    if user_message is not None:
        run_messages.user_message = user_message
        run_messages.messages.append(user_message)

    # Set messages on run_context so tool hooks can access the current message history
    run_context.messages = run_messages.messages

    return run_messages


def _get_user_message(
    team: "Team",
    *,
    run_response: TeamRunOutput,
    run_context: RunContext,
    input_message: Optional[Union[str, List, Dict, Message, BaseModel, List[Message]]] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    add_dependencies_to_context: Optional[bool] = None,
    **kwargs,
):
    # Get references from the knowledge base to use in the user message
    from agno.team._utils import _convert_dependencies_to_string, _convert_documents_to_string

    references = None

    if input_message is None:
        # If we have any media, return a message with empty content
        if images is not None or audio is not None or videos is not None or files is not None:
            return Message(
                role="user",
                content="",
                images=None if not team.send_media_to_model else images,
                audio=None if not team.send_media_to_model else audio,
                videos=None if not team.send_media_to_model else videos,
                files=None if not team.send_media_to_model else files,
                **kwargs,
            )
        else:
            # If the input is None, return None
            return None

    else:
        if isinstance(input_message, list):
            input_content: Union[str, List[Any], List[Message]]
            if len(input_message) > 0 and isinstance(input_message[0], dict) and "type" in input_message[0]:
                # This is multimodal content (text + images/audio/video), preserve the structure
                input_content = input_message
            elif len(input_message) > 0 and isinstance(input_message[0], Message):
                # This is a list of Message objects, extract text content from them
                input_content = get_text_from_message(input_message)
            elif all(isinstance(item, str) for item in input_message):
                input_content = "\n".join([str(item) for item in input_message])
            else:
                input_content = str(input_message)

            return Message(
                role="user",
                content=input_content,
                images=None if not team.send_media_to_model else images,
                audio=None if not team.send_media_to_model else audio,
                videos=None if not team.send_media_to_model else videos,
                files=None if not team.send_media_to_model else files,
                **kwargs,
            )

        # If message is provided as a Message, use it directly
        elif isinstance(input_message, Message):
            return input_message
        # If message is provided as a dict, try to validate it as a Message
        elif isinstance(input_message, dict):
            try:
                if team.input_schema and is_typed_dict(team.input_schema):
                    import json

                    content = json.dumps(input_message, indent=2, ensure_ascii=False)
                    return Message(role="user", content=content)
                else:
                    return Message.model_validate(input_message)
            except Exception as e:
                log_warning(f"Failed to validate input: {e}")

        # If message is provided as a BaseModel, convert it to a Message
        elif isinstance(input_message, BaseModel):
            try:
                # Create a user message with the BaseModel content
                content = input_message.model_dump_json(indent=2, exclude_none=True)
                return Message(role="user", content=content)
            except Exception as e:
                log_warning(f"Failed to convert BaseModel to message: {e}")
        else:
            user_msg_content = input_message
            if team.add_knowledge_to_context:
                if isinstance(input_message, str):
                    user_msg_content = input_message
                elif callable(input_message):
                    user_msg_content = input_message(agent=team)
                else:
                    raise Exception("input must be a string or a callable when add_references is True")

                try:
                    retrieval_timer = Timer()
                    retrieval_timer.start()
                    docs_from_knowledge = team.get_relevant_docs_from_knowledge(
                        query=user_msg_content,
                        filters=run_context.knowledge_filters,
                        run_context=run_context,
                        **kwargs,
                    )
                    if docs_from_knowledge is not None:
                        references = MessageReferences(
                            query=user_msg_content,
                            references=docs_from_knowledge,
                            time=round(retrieval_timer.elapsed, 4),
                        )
                        # Add the references to the run_response
                        if run_response.references is None:
                            run_response.references = []
                        run_response.references.append(references)
                    retrieval_timer.stop()
                    log_debug(f"Time to get references: {retrieval_timer.elapsed:.4f}s")
                except Exception as e:
                    log_warning(f"Failed to get references: {e}")

            if team.resolve_in_context:
                user_msg_content = _format_message_with_state_variables(
                    team,
                    user_msg_content,
                    run_context=run_context,
                )

            # Convert to string for concatenation operations
            user_msg_content_str = get_text_from_message(user_msg_content) if user_msg_content is not None else ""

            # 4.1 Add knowledge references to user message
            if (
                team.add_knowledge_to_context
                and references is not None
                and references.references is not None
                and len(references.references) > 0
            ):
                user_msg_content_str += "\n\nUse the following references from the knowledge base if it helps:\n"
                user_msg_content_str += "<references>\n"
                user_msg_content_str += _convert_documents_to_string(team, references.references) + "\n"
                user_msg_content_str += "</references>"
            # 4.2 Add context to user message
            if add_dependencies_to_context and run_context.dependencies is not None:
                user_msg_content_str += "\n\n<additional context>\n"
                user_msg_content_str += _convert_dependencies_to_string(team, run_context.dependencies) + "\n"
                user_msg_content_str += "</additional context>"

            # Use the string version for the final content
            user_msg_content = user_msg_content_str

            # Return the user message
            return Message(
                role="user",
                content=user_msg_content,
                images=None if not team.send_media_to_model else images,
                audio=None if not team.send_media_to_model else audio,
                videos=None if not team.send_media_to_model else videos,
                files=None if not team.send_media_to_model else files,
                **kwargs,
            )


async def _aget_user_message(
    team: "Team",
    *,
    run_response: TeamRunOutput,
    run_context: RunContext,
    input_message: Optional[Union[str, List, Dict, Message, BaseModel, List[Message]]] = None,
    audio: Optional[Sequence[Audio]] = None,
    images: Optional[Sequence[Image]] = None,
    videos: Optional[Sequence[Video]] = None,
    files: Optional[Sequence[File]] = None,
    add_dependencies_to_context: Optional[bool] = None,
    **kwargs,
):
    # Get references from the knowledge base to use in the user message
    from agno.team._utils import _convert_dependencies_to_string, _convert_documents_to_string

    references = None

    if input_message is None:
        # If we have any media, return a message with empty content
        if images is not None or audio is not None or videos is not None or files is not None:
            return Message(
                role="user",
                content="",
                images=None if not team.send_media_to_model else images,
                audio=None if not team.send_media_to_model else audio,
                videos=None if not team.send_media_to_model else videos,
                files=None if not team.send_media_to_model else files,
                **kwargs,
            )
        else:
            # If the input is None, return None
            return None

    else:
        if isinstance(input_message, list):
            input_content: Union[str, List[Any], List[Message]]
            if len(input_message) > 0 and isinstance(input_message[0], dict) and "type" in input_message[0]:
                # This is multimodal content (text + images/audio/video), preserve the structure
                input_content = input_message
            elif len(input_message) > 0 and isinstance(input_message[0], Message):
                # This is a list of Message objects, extract text content from them
                input_content = get_text_from_message(input_message)
            elif all(isinstance(item, str) for item in input_message):
                input_content = "\n".join([str(item) for item in input_message])
            else:
                input_content = str(input_message)

            return Message(
                role="user",
                content=input_content,
                images=None if not team.send_media_to_model else images,
                audio=None if not team.send_media_to_model else audio,
                videos=None if not team.send_media_to_model else videos,
                files=None if not team.send_media_to_model else files,
                **kwargs,
            )

        # If message is provided as a Message, use it directly
        elif isinstance(input_message, Message):
            return input_message
        # If message is provided as a dict, try to validate it as a Message
        elif isinstance(input_message, dict):
            try:
                if team.input_schema and is_typed_dict(team.input_schema):
                    import json

                    content = json.dumps(input_message, indent=2, ensure_ascii=False)
                    return Message(role="user", content=content)
                else:
                    return Message.model_validate(input_message)
            except Exception as e:
                log_warning(f"Failed to validate input: {e}")

        # If message is provided as a BaseModel, convert it to a Message
        elif isinstance(input_message, BaseModel):
            try:
                # Create a user message with the BaseModel content
                content = input_message.model_dump_json(indent=2, exclude_none=True)
                return Message(role="user", content=content)
            except Exception as e:
                log_warning(f"Failed to convert BaseModel to message: {e}")
        else:
            user_msg_content = input_message
            if team.add_knowledge_to_context:
                if isinstance(input_message, str):
                    user_msg_content = input_message
                elif callable(input_message):
                    user_msg_content = input_message(agent=team)
                else:
                    raise Exception("input must be a string or a callable when add_references is True")

                try:
                    retrieval_timer = Timer()
                    retrieval_timer.start()
                    docs_from_knowledge = await team.aget_relevant_docs_from_knowledge(
                        query=user_msg_content,
                        filters=run_context.knowledge_filters,
                        run_context=run_context,
                        **kwargs,
                    )
                    if docs_from_knowledge is not None:
                        references = MessageReferences(
                            query=user_msg_content,
                            references=docs_from_knowledge,
                            time=round(retrieval_timer.elapsed, 4),
                        )
                        # Add the references to the run_response
                        if run_response.references is None:
                            run_response.references = []
                        run_response.references.append(references)
                    retrieval_timer.stop()
                    log_debug(f"Time to get references: {retrieval_timer.elapsed:.4f}s")
                except Exception as e:
                    log_warning(f"Failed to get references: {e}")

            if team.resolve_in_context:
                user_msg_content = _format_message_with_state_variables(
                    team,
                    user_msg_content,
                    run_context=run_context,
                )

            # Convert to string for concatenation operations
            user_msg_content_str = get_text_from_message(user_msg_content) if user_msg_content is not None else ""

            # 4.1 Add knowledge references to user message
            if (
                team.add_knowledge_to_context
                and references is not None
                and references.references is not None
                and len(references.references) > 0
            ):
                user_msg_content_str += "\n\nUse the following references from the knowledge base if it helps:\n"
                user_msg_content_str += "<references>\n"
                user_msg_content_str += _convert_documents_to_string(team, references.references) + "\n"
                user_msg_content_str += "</references>"
            # 4.2 Add context to user message
            if add_dependencies_to_context and run_context.dependencies is not None:
                user_msg_content_str += "\n\n<additional context>\n"
                user_msg_content_str += _convert_dependencies_to_string(team, run_context.dependencies) + "\n"
                user_msg_content_str += "</additional context>"

            # Use the string version for the final content
            user_msg_content = user_msg_content_str

            # Return the user message
            return Message(
                role="user",
                content=user_msg_content,
                images=None if not team.send_media_to_model else images,
                audio=None if not team.send_media_to_model else audio,
                videos=None if not team.send_media_to_model else videos,
                files=None if not team.send_media_to_model else files,
                **kwargs,
            )


def _get_messages_for_parser_model(
    team: "Team",
    model_response: ModelResponse,
    response_format: Optional[Union[Dict, Type[BaseModel]]],
    run_context: Optional[RunContext] = None,
) -> List[Message]:
    """Get the messages for the parser model."""
    from agno.utils.prompts import get_json_output_prompt

    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    system_content = (
        team.parser_model_prompt
        if team.parser_model_prompt is not None
        else "You are tasked with creating a structured output from the provided user message."
    )

    if response_format == {"type": "json_object"} and output_schema is not None:
        system_content += f"{get_json_output_prompt(output_schema)}"  # type: ignore

    return [
        Message(role="system", content=system_content),
        Message(role="user", content=model_response.content),
    ]


def _get_messages_for_parser_model_stream(
    team: "Team",
    run_response: TeamRunOutput,
    response_format: Optional[Union[Dict, Type[BaseModel]]],
    run_context: Optional[RunContext] = None,
) -> List[Message]:
    """Get the messages for the parser model."""
    from agno.utils.prompts import get_json_output_prompt

    # Get output_schema from run_context
    output_schema = run_context.output_schema if run_context else None

    system_content = (
        team.parser_model_prompt
        if team.parser_model_prompt is not None
        else "You are tasked with creating a structured output from the provided data."
    )

    if response_format == {"type": "json_object"} and output_schema is not None:
        system_content += f"{get_json_output_prompt(output_schema)}"  # type: ignore

    return [
        Message(role="system", content=system_content),
        Message(role="user", content=run_response.content),
    ]


def _get_messages_for_output_model(team: "Team", messages: List[Message]) -> List[Message]:
    """Get the messages for the output model."""
    from copy import deepcopy

    # Copy the list and messages to avoid mutating the originals
    messages = [deepcopy(m) for m in messages]

    if team.output_model_prompt is not None:
        system_message_exists = False
        for message in messages:
            if message.role == "system":
                system_message_exists = True
                message.content = team.output_model_prompt
                break
        if not system_message_exists:
            messages.insert(0, Message(role="system", content=team.output_model_prompt))

    # Remove the last assistant message from the messages list
    if messages and messages[-1].role == "assistant":
        messages.pop(-1)

    return messages


def _format_message_with_state_variables(
    team: "Team",
    message: Any,
    run_context: Optional[RunContext] = None,
) -> Any:
    """Format a message with the session state variables from run_context."""
    import re
    import string

    if not isinstance(message, str):
        return message

    # Extract values from run_context
    session_state = run_context.session_state if run_context else None
    dependencies = run_context.dependencies if run_context else None
    metadata = run_context.metadata if run_context else None
    user_id = run_context.user_id if run_context else None

    # Should already be resolved and passed from run() method
    format_variables = ChainMap(
        session_state if session_state is not None else {},
        dependencies or {},
        metadata or {},
        {"user_id": user_id} if user_id is not None else {},
    )
    converted_msg = message
    for var_name in format_variables.keys():
        # Only convert standalone {var_name} patterns, not nested ones
        pattern = r"\{" + re.escape(var_name) + r"\}"
        replacement = "${" + var_name + "}"
        converted_msg = re.sub(pattern, replacement, converted_msg)

    # Use Template to safely substitute variables
    template = string.Template(converted_msg)
    try:
        result = template.safe_substitute(format_variables)
        return result
    except Exception as e:
        log_warning(f"Template substitution failed: {e}")
        return message


def _get_json_output_prompt(
    team: "Team", output_schema: Optional[Union[Type[BaseModel], Dict[str, Any]]] = None
) -> str:
    """Return the JSON output prompt for the Agent.

    This is added to the system prompt when the output_schema is set and structured_outputs is False.
    """

    json_output_prompt = "Provide your output as a JSON containing the following fields:"
    if output_schema is not None:
        if isinstance(output_schema, str):
            json_output_prompt += "\n<json_fields>"
            json_output_prompt += f"\n{output_schema}"
            json_output_prompt += "\n</json_fields>"
        elif isinstance(output_schema, list):
            json_output_prompt += "\n<json_fields>"
            json_output_prompt += f"\n{json.dumps(output_schema)}"
            json_output_prompt += "\n</json_fields>"
        elif isinstance(output_schema, dict):
            json_output_prompt += "\n<json_fields>"
            json_output_prompt += f"\n{json.dumps(output_schema)}"
            json_output_prompt += "\n</json_fields>"
        elif isinstance(output_schema, type) and issubclass(output_schema, BaseModel):
            json_schema = output_schema.model_json_schema()
            if json_schema is not None:
                response_model_properties = {}
                json_schema_properties = json_schema.get("properties")
                if json_schema_properties is not None:
                    for field_name, field_properties in json_schema_properties.items():
                        formatted_field_properties = {
                            prop_name: prop_value
                            for prop_name, prop_value in field_properties.items()
                            if prop_name != "title"
                        }
                        response_model_properties[field_name] = formatted_field_properties
                json_schema_defs = json_schema.get("$defs")
                if json_schema_defs is not None:
                    response_model_properties["$defs"] = {}
                    for def_name, def_properties in json_schema_defs.items():
                        def_fields = def_properties.get("properties")
                        formatted_def_properties = {}
                        if def_fields is not None:
                            for field_name, field_properties in def_fields.items():
                                formatted_field_properties = {
                                    prop_name: prop_value
                                    for prop_name, prop_value in field_properties.items()
                                    if prop_name != "title"
                                }
                                formatted_def_properties[field_name] = formatted_field_properties
                        if len(formatted_def_properties) > 0:
                            response_model_properties["$defs"][def_name] = formatted_def_properties

                if len(response_model_properties) > 0:
                    json_output_prompt += "\n<json_fields>"
                    json_output_prompt += (
                        f"\n{json.dumps([key for key in response_model_properties.keys() if key != '$defs'])}"
                    )
                    json_output_prompt += "\n</json_fields>"
                    json_output_prompt += "\n\nHere are the properties for each field:"
                    json_output_prompt += "\n<json_field_properties>"
                    json_output_prompt += f"\n{json.dumps(response_model_properties, indent=2)}"
                    json_output_prompt += "\n</json_field_properties>"
        else:
            log_warning(f"Could not build json schema for {output_schema}")
    else:
        json_output_prompt += "Provide the output as JSON."

    json_output_prompt += "\nStart your response with `{` and end it with `}`."
    json_output_prompt += "\nYour output will be passed to json.loads() to convert it to a Python object."
    json_output_prompt += "\nMake sure it only contains valid JSON."
    return json_output_prompt



import hashlib
import hmac
import os
from typing import Optional

from fastapi import HTTPException

from agno.utils.log import log_warning


def validate_webhook_signature(payload: bytes, signature_header: Optional[str]) -> bool:
    app_secret = os.getenv("WHATSAPP_APP_SECRET")
    if not app_secret:
        # Explicit opt-out: operator must deliberately set this for local dev
        if os.getenv("WHATSAPP_SKIP_SIGNATURE_VALIDATION", "").lower() == "true":
            log_warning("WHATSAPP_SKIP_SIGNATURE_VALIDATION=true — signature check disabled")
            return True
        raise HTTPException(
            status_code=500,
            detail="WHATSAPP_APP_SECRET not set. Set WHATSAPP_SKIP_SIGNATURE_VALIDATION=true for local development.",
        )

    if not signature_header or not signature_header.startswith("sha256="):
        return False

    # Header format: "sha256=<hex>"; strip prefix to compare digests
    expected_signature = signature_header.removeprefix("sha256=")

    hmac_obj = hmac.new(app_secret.encode(), payload, hashlib.sha256)
    calculated_signature = hmac_obj.hexdigest()

    # Constant-time comparison prevents timing side-channels
    return hmac.compare_digest(calculated_signature, expected_signature)



import json
import os
from functools import wraps
from typing import Any, Dict, List, Literal, Optional, Set
from urllib.parse import urlencode

from agno.tools import Toolkit
from agno.utils.log import log_error


def google_authenticate(service_name: str):
    """Shared auth decorator for all Google toolkits.

    Each toolkit creates a module-level alias:
        authenticate = google_authenticate("gmail")

    Expects the toolkit class to define:
        - self.creds: Google OAuth credentials
        - self.service: Built API client (set by _build_service)
        - self._auth(): Loads or refreshes credentials
        - self._build_service(): Returns build(api_name, api_version, credentials=self.creds)
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.creds or not self.creds.valid:
                try:
                    self._auth()
                except Exception as e:
                    log_error(f"{service_name.title()} authentication failed: {e}")
                    return json.dumps({"error": f"{service_name.title()} authentication failed: {e}"})
            if not self.service:
                try:
                    self.service = self._build_service()
                except Exception as e:
                    log_error(f"{service_name.title()} service initialization failed: {e}")
                    return json.dumps({"error": f"{service_name.title()} service initialization failed: {e}"})
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


class GoogleAuth(Toolkit):
    def __init__(
        self,
        client_id: Optional[str] = None,
        redirect_uri: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(
            name="google_auth",
            instructions="When any Google tool (Gmail, Calendar, Drive, Sheets) returns an authentication error, immediately call authenticate_google to get the OAuth URL for the user.",
            **kwargs,
        )
        self.client_id = client_id or os.getenv("GOOGLE_CLIENT_ID")
        self.redirect_uri = redirect_uri or os.getenv("GOOGLE_REDIRECT_URI", "http://localhost:8080/")
        self._services: Dict[str, List[str]] = {}
        self.register(self.authenticate_google)

    def register_service(self, service: str, scopes: List[str]) -> None:
        self._services[service] = scopes

    def authenticate_google(self, services: List[Literal["gmail", "calendar", "drive", "sheets", "slides"]]) -> str:
        """
        Get the Google OAuth URL for the user to authenticate their Google account.

        Args:
            services (List[str]): Google services to authenticate.

        Returns:
            str: JSON string containing the OAuth URL or error message
        """
        scopes: Set[str] = set()
        for service in services:
            if service in self._services:
                scopes.update(self._services[service])
        if not scopes:
            return json.dumps({"error": f"Unknown services. Available: {', '.join(self._services)}"})
        params = {
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "scope": " ".join(scopes),
            "response_type": "code",
            "access_type": "offline",
            "prompt": "consent",
            "include_granted_scopes": "true",
        }
        url = f"https://accounts.google.com/o/oauth2/v2/auth?{urlencode(params)}"
        return json.dumps({"message": f"Connect {', '.join(services)}", "url": url})



from agno.eval.base import BaseEval

__all__ = [
    "AccuracyAgentResponse",
    "AccuracyEvaluation",
    "AccuracyResult",
    "AccuracyEval",
    "AgentAsJudgeEval",
    "AgentAsJudgeEvaluation",
    "AgentAsJudgeResult",
    "BaseEval",
    "PerformanceEval",
    "PerformanceResult",
    "ReliabilityEval",
    "ReliabilityResult",
]


def __getattr__(name: str):
    """Lazy import for eval implementations to avoid circular imports with Agent."""
    if name in ("AccuracyAgentResponse", "AccuracyEval", "AccuracyEvaluation", "AccuracyResult"):
        from agno.eval import accuracy

        return getattr(accuracy, name)
    elif name in ("AgentAsJudgeEval", "AgentAsJudgeEvaluation", "AgentAsJudgeResult"):
        from agno.eval import agent_as_judge

        return getattr(agent_as_judge, name)
    elif name in ("PerformanceEval", "PerformanceResult"):
        from agno.eval import performance

        return getattr(performance, name)
    elif name in ("ReliabilityEval", "ReliabilityResult"):
        from agno.eval import reliability

        return getattr(reliability, name)
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from typing import Any, Dict, Optional
from uuid import uuid4

from pydantic import BaseModel

from agno.agent import Agent
from agno.models.message import Message
from agno.os.schema import ModelResponse
from agno.os.utils import (
    format_tools,
)
from agno.run import RunContext
from agno.run.agent import RunOutput
from agno.session import AgentSession
from agno.utils.agent import aexecute_instructions, aexecute_system_message


class AgentResponse(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    db_id: Optional[str] = None
    description: Optional[str] = None
    role: Optional[str] = None
    model: Optional[ModelResponse] = None
    tools: Optional[Dict[str, Any]] = None
    sessions: Optional[Dict[str, Any]] = None
    knowledge: Optional[Dict[str, Any]] = None
    memory: Optional[Dict[str, Any]] = None
    reasoning: Optional[Dict[str, Any]] = None
    default_tools: Optional[Dict[str, Any]] = None
    system_message: Optional[Dict[str, Any]] = None
    extra_messages: Optional[Dict[str, Any]] = None
    response_settings: Optional[Dict[str, Any]] = None
    introduction: Optional[str] = None
    streaming: Optional[Dict[str, Any]] = None
    metadata: Optional[Dict[str, Any]] = None
    input_schema: Optional[Dict[str, Any]] = None
    is_component: bool = False
    current_version: Optional[int] = None
    stage: Optional[str] = None

    @classmethod
    async def from_agent(
        cls,
        agent: Agent,
        is_component: bool = False,
    ) -> "AgentResponse":
        def filter_meaningful_config(d: Dict[str, Any], defaults: Dict[str, Any]) -> Optional[Dict[str, Any]]:
            """Filter out fields that match their default values, keeping only meaningful user configurations"""
            filtered = {}
            for key, value in d.items():
                if value is None:
                    continue
                # Skip if value matches the default exactly
                if key in defaults and value == defaults[key]:
                    continue
                # Keep non-default values
                filtered[key] = value
            return filtered if filtered else None

        # Define default values for filtering
        agent_defaults = {
            # Sessions defaults
            "add_history_to_context": False,
            "num_history_runs": 3,
            "enable_session_summaries": False,
            "search_past_sessions": False,
            "cache_session": False,
            # Knowledge defaults
            "add_references": False,
            "references_format": "json",
            "enable_agentic_knowledge_filters": False,
            # Memory defaults
            "enable_agentic_memory": False,
            "update_memory_on_run": False,
            # Reasoning defaults
            "reasoning": False,
            "reasoning_min_steps": 1,
            "reasoning_max_steps": 10,
            # Default tools defaults
            "read_chat_history": False,
            "search_knowledge": True,
            "update_knowledge": False,
            "read_tool_call_history": False,
            # System message defaults
            "system_message_role": "system",
            "build_context": True,
            "markdown": False,
            "add_name_to_context": False,
            "add_datetime_to_context": False,
            "add_location_to_context": False,
            "resolve_in_context": True,
            # Extra messages defaults
            "user_message_role": "user",
            "build_user_context": True,
            # Response settings defaults
            "retries": 0,
            "delay_between_retries": 1,
            "exponential_backoff": False,
            "parse_response": True,
            "use_json_mode": False,
            # Streaming defaults
            "stream_events": False,
        }

        session_id = str(uuid4())
        run_id = str(uuid4())
        agent_tools = await agent.aget_tools(
            session=AgentSession(session_id=session_id, session_data={}),
            run_response=RunOutput(run_id=run_id, session_id=session_id),
            run_context=RunContext(run_id=run_id, session_id=session_id, user_id=agent.user_id),
            check_mcp_tools=False,
        )
        formatted_tools = format_tools(agent_tools) if agent_tools else None

        additional_input = agent.additional_input
        if additional_input and isinstance(additional_input[0], Message):
            additional_input = [message.to_dict() for message in additional_input]  # type: ignore

        input_schema_dict = None
        if agent.input_schema is not None:
            if isinstance(agent.input_schema, dict):
                input_schema_dict = agent.input_schema
            else:
                try:
                    input_schema_dict = agent.input_schema.model_json_schema()
                except Exception:
                    pass

        # Build model only if it has at least one non-null field
        model_name = agent.model.name if (agent.model and agent.model.name) else None
        model_provider = agent.model.provider if (agent.model and agent.model.provider) else None
        model_id = agent.model.id if (agent.model and agent.model.id) else None
        _agent_model_data: Dict[str, Any] = {}
        if model_name is not None:
            _agent_model_data["name"] = model_name
        if model_id is not None:
            _agent_model_data["model"] = model_id
        if model_provider is not None:
            _agent_model_data["provider"] = model_provider

        session_table = agent.db.session_table_name if agent.db else None
        knowledge_table = agent.db.knowledge_table_name if agent.db and agent.knowledge else None

        tools_info = {
            "tools": formatted_tools,
            "tool_call_limit": agent.tool_call_limit,
            "tool_choice": agent.tool_choice,
        }

        sessions_info = {
            "session_table": session_table,
            "add_history_to_context": agent.add_history_to_context,
            "enable_session_summaries": agent.enable_session_summaries,
            "num_history_runs": agent.num_history_runs,
            "search_past_sessions": agent.search_past_sessions,
            "num_past_sessions_to_search": agent.num_past_sessions_to_search,
            "num_past_session_runs_in_search": agent.num_past_session_runs_in_search,
            "cache_session": agent.cache_session,
        }

        contents_db = getattr(agent.knowledge, "contents_db", None) if agent.knowledge else None
        knowledge_info = {
            "db_id": contents_db.id if contents_db else None,
            "knowledge_table": knowledge_table,
            "enable_agentic_knowledge_filters": agent.enable_agentic_knowledge_filters,
            "knowledge_filters": (
                [f.to_dict() if hasattr(f, "to_dict") else f for f in agent.knowledge_filters]
                if isinstance(agent.knowledge_filters, list)
                else agent.knowledge_filters
            ),
            "references_format": agent.references_format,
        }

        memory_info: Optional[Dict[str, Any]] = None
        if agent.memory_manager is not None:
            memory_info = {
                "enable_agentic_memory": agent.enable_agentic_memory,
                "update_memory_on_run": agent.update_memory_on_run,
                "enable_user_memories": agent.enable_user_memories,  # Soon to be deprecated. Use update_memory_on_run
                "metadata": agent.metadata,
                "memory_table": agent.db.memory_table_name if agent.db and agent.update_memory_on_run else None,
            }

            if agent.memory_manager.model is not None:
                memory_info["model"] = ModelResponse(
                    name=agent.memory_manager.model.name,
                    model=agent.memory_manager.model.id,
                    provider=agent.memory_manager.model.provider,
                ).model_dump()

        reasoning_info: Dict[str, Any] = {
            "reasoning": agent.reasoning,
            "reasoning_agent_id": agent.reasoning_agent.id if agent.reasoning_agent else None,
            "reasoning_min_steps": agent.reasoning_min_steps,
            "reasoning_max_steps": agent.reasoning_max_steps,
        }

        if agent.reasoning_model:
            reasoning_info["reasoning_model"] = ModelResponse(
                name=agent.reasoning_model.name,
                model=agent.reasoning_model.id,
                provider=agent.reasoning_model.provider,
            ).model_dump()

        default_tools_info = {
            "read_chat_history": agent.read_chat_history,
            "search_knowledge": agent.search_knowledge,
            "update_knowledge": agent.update_knowledge,
            "read_tool_call_history": agent.read_tool_call_history,
        }

        instructions = agent.instructions if agent.instructions else None
        if instructions and callable(instructions):
            instructions = await aexecute_instructions(instructions=instructions, agent=agent)

        system_message = agent.system_message if agent.system_message else None
        if system_message and callable(system_message):
            system_message = await aexecute_system_message(system_message=system_message, agent=agent)

        system_message_info = {
            "system_message": str(system_message) if system_message else None,
            "system_message_role": agent.system_message_role,
            "build_context": agent.build_context,
            "description": agent.description,
            "instructions": instructions,
            "expected_output": agent.expected_output,
            "additional_context": agent.additional_context,
            "markdown": agent.markdown,
            "add_name_to_context": agent.add_name_to_context,
            "add_datetime_to_context": agent.add_datetime_to_context,
            "add_location_to_context": agent.add_location_to_context,
            "timezone_identifier": agent.timezone_identifier,
            "resolve_in_context": agent.resolve_in_context,
        }

        extra_messages_info = {
            "additional_input": additional_input,  # type: ignore
            "user_message_role": agent.user_message_role,
            "build_user_context": agent.build_user_context,
        }

        # Handle output_schema name for both Pydantic models and JSON schemas
        output_schema_name = None
        if agent.output_schema is not None:
            if isinstance(agent.output_schema, dict):
                if "json_schema" in agent.output_schema:
                    output_schema_name = agent.output_schema["json_schema"].get("name", "JSONSchema")
                elif "schema" in agent.output_schema and isinstance(agent.output_schema["schema"], dict):
                    output_schema_name = agent.output_schema["schema"].get("title", "JSONSchema")
                else:
                    output_schema_name = agent.output_schema.get("title", "JSONSchema")
            elif hasattr(agent.output_schema, "__name__"):
                output_schema_name = agent.output_schema.__name__

        response_settings_info: Dict[str, Any] = {
            "retries": agent.retries,
            "delay_between_retries": agent.delay_between_retries,
            "exponential_backoff": agent.exponential_backoff,
            "output_schema_name": output_schema_name,
            "parser_model_prompt": agent.parser_model_prompt,
            "parse_response": agent.parse_response,
            "structured_outputs": agent.structured_outputs,
            "use_json_mode": agent.use_json_mode,
            "save_response_to_file": agent.save_response_to_file,
        }

        if agent.parser_model:
            response_settings_info["parser_model"] = ModelResponse(
                name=agent.parser_model.name,
                model=agent.parser_model.id,
                provider=agent.parser_model.provider,
            ).model_dump()

        streaming_info = {
            "stream": agent.stream,
            "stream_events": agent.stream_events,
        }

        return AgentResponse(
            id=agent.id,
            name=agent.name,
            db_id=agent.db.id if agent.db else None,
            description=agent.description,
            role=agent.role,
            model=ModelResponse(**_agent_model_data) if _agent_model_data else None,
            tools=filter_meaningful_config(tools_info, {}),
            sessions=filter_meaningful_config(sessions_info, agent_defaults),
            knowledge=filter_meaningful_config(knowledge_info, agent_defaults),
            memory=filter_meaningful_config(memory_info, agent_defaults) if memory_info else None,
            reasoning=filter_meaningful_config(reasoning_info, agent_defaults),
            default_tools=filter_meaningful_config(default_tools_info, agent_defaults),
            system_message=filter_meaningful_config(system_message_info, agent_defaults),
            extra_messages=filter_meaningful_config(extra_messages_info, agent_defaults),
            response_settings=filter_meaningful_config(response_settings_info, agent_defaults),
            streaming=filter_meaningful_config(streaming_info, agent_defaults),
            introduction=agent.introduction,
            metadata=agent.metadata,
            input_schema=input_schema_dict,
            is_component=is_component,
            current_version=getattr(agent, "_version", None),
            stage=getattr(agent, "_stage", None),
        )
